from neuron import h, nrn, nrn_dll_sym
from . import species, node, section1d, region, generalizedReaction, constants
from .nodelist import NodeList
from .node import _point_indices
import weakref
import numpy
import ctypes
import atexit
from . import options
from .rxdException import RxDException
from . import initializer
import collections
import os
import sysconfig
import uuid
import sys
import itertools
from numpy.ctypeslib import ndpointer
import re
import platform
from warnings import warn

# aliases to avoid repeatedly doing multiple hash-table lookups
_numpy_array = numpy.array
_numpy_zeros = numpy.zeros
_species_get_all_species = species._get_all_species
_node_get_states = node._get_states
_section1d_transfer_to_legacy = section1d._transfer_to_legacy
_ctypes_c_int = ctypes.c_int
_weakref_ref = weakref.ref

_external_solver = None
_external_solver_initialized = False
_windows_dll_files = []
_windows_dll = []


make_time_ptr = nrn_dll_sym("make_time_ptr")
make_time_ptr.argtypes = [ctypes.py_object, ctypes.py_object]
make_time_ptr(h._ref_dt, h._ref_t)

_double_ptr = ctypes.POINTER(ctypes.c_double)
_int_ptr = ctypes.POINTER(_ctypes_c_int)
_long_ptr = ctypes.POINTER(ctypes.c_long)


fptr_prototype = ctypes.CFUNCTYPE(None)
set_nonvint_block = nrn_dll_sym("set_nonvint_block")
set_nonvint_block(nrn_dll_sym("rxd_nonvint_block"))

set_setup = nrn_dll_sym("set_setup")
set_setup.argtypes = [fptr_prototype]
set_initialize = nrn_dll_sym("set_initialize")
set_initialize.argtypes = [fptr_prototype]

scatter_concentrations = nrn_dll_sym("scatter_concentrations")

# Transfer extracellular concentrations to NEURON
_fih_transfer_ecs = h.FInitializeHandler(1, scatter_concentrations)


rxd_set_no_diffusion = nrn_dll_sym("rxd_set_no_diffusion")

setup_solver = nrn_dll_sym("setup_solver")
setup_solver.argtypes = [
    ndpointer(ctypes.c_double),
    ctypes.c_int,
    numpy.ctypeslib.ndpointer(numpy.int_, flags="contiguous"),
    ctypes.c_int,
]

# states = None
_set_num_threads = nrn_dll_sym("set_num_threads")
_set_num_threads.argtypes = [ctypes.c_int]
_get_num_threads = nrn_dll_sym("get_num_threads")
_get_num_threads.restype = ctypes.c_int

free_conc_ptrs = nrn_dll_sym("free_conc_ptrs")
free_curr_ptrs = nrn_dll_sym("free_curr_ptrs")
clear_rates = nrn_dll_sym("clear_rates")
register_rate = nrn_dll_sym("register_rate")
register_rate.argtypes = [
    ctypes.c_int,  # num species
    ctypes.c_int,  # num parameters
    ctypes.c_int,  # num regions
    ctypes.c_int,  # num seg
    numpy.ctypeslib.ndpointer(ctypes.c_int, flags="contiguous"),  # species ids
    ctypes.c_int,  # num ecs species
    ctypes.c_int,  # num ecs parameters
    numpy.ctypeslib.ndpointer(ctypes.c_int, flags="contiguous"),  # ecs species ids
    numpy.ctypeslib.ndpointer(ctypes.c_int, flags="contiguous"),  # ecs indices
    ctypes.c_int,  # num multicompartment reactions
    numpy.ctypeslib.ndpointer(
        ctypes.c_double, flags="contiguous"
    ),  # multicompartment multipliers
    ctypes.POINTER(ctypes.py_object),  # voltage pointers
]  # Reaction rate function

setup_currents = nrn_dll_sym("setup_currents")
setup_currents.argtypes = [
    ctypes.c_int,  # number of membrane currents
    ctypes.c_int,  # number induced currents
    _int_ptr,  # number of species involved in each membrane current
    _int_ptr,  # node indices
    _double_ptr,  # scaling (areas) of the fluxes
    ctypes.POINTER(ctypes.py_object),  # hoc pointers
    _int_ptr,  # maps for membrane fluxes
    _int_ptr,  # maps for ecs fluxes
]


ics_register_reaction = nrn_dll_sym("ics_register_reaction")
ics_register_reaction.argtypes = [
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
    _int_ptr,
    numpy.ctypeslib.ndpointer(dtype=numpy.uint64),
    ctypes.c_int,
    numpy.ctypeslib.ndpointer(dtype=float),
]

ecs_register_reaction = nrn_dll_sym("ecs_register_reaction")
ecs_register_reaction.argtypes = [
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
    _int_ptr,
]


set_hybrid_data = nrn_dll_sym("set_hybrid_data")
set_hybrid_data.argtypes = [
    numpy.ctypeslib.ndpointer(dtype=numpy.int64),
    numpy.ctypeslib.ndpointer(dtype=numpy.int64),
    numpy.ctypeslib.ndpointer(dtype=numpy.int64),
    numpy.ctypeslib.ndpointer(dtype=numpy.int64),
    numpy.ctypeslib.ndpointer(dtype=numpy.int64),
    numpy.ctypeslib.ndpointer(dtype=numpy.int64),
    numpy.ctypeslib.ndpointer(dtype=float),
    numpy.ctypeslib.ndpointer(dtype=float),
    numpy.ctypeslib.ndpointer(dtype=float),
    numpy.ctypeslib.ndpointer(dtype=float),
]

# ics_register_reaction = nrn_dll_sym('ics_register_reaction')
# ics_register_reaction.argtype = [ctypes.c_int, ctypes.c_int, _int_ptr, fptr_prototype]

set_euler_matrix = nrn_dll_sym("rxd_set_euler_matrix")
set_euler_matrix.argtypes = [
    ctypes.c_int,
    ctypes.c_int,
    _long_ptr,
    _long_ptr,
    _double_ptr,
    numpy.ctypeslib.ndpointer(numpy.double, flags="contiguous"),
]
rxd_setup_curr_ptrs = nrn_dll_sym("rxd_setup_curr_ptrs")
rxd_setup_curr_ptrs.argtypes = [
    ctypes.c_int,
    _int_ptr,
    numpy.ctypeslib.ndpointer(numpy.double, flags="contiguous"),
    ctypes.POINTER(ctypes.py_object),
]

rxd_setup_conc_ptrs = nrn_dll_sym("rxd_setup_conc_ptrs")
rxd_setup_conc_ptrs.argtypes = [
    ctypes.c_int,
    _int_ptr,
    ctypes.POINTER(ctypes.py_object),
]

rxd_include_node_flux1D = nrn_dll_sym("rxd_include_node_flux1D")
rxd_include_node_flux1D.argtypes = [
    ctypes.c_int,
    _long_ptr,
    _double_ptr,
    ctypes.POINTER(ctypes.py_object),
]

rxd_include_node_flux3D = nrn_dll_sym("rxd_include_node_flux3D")
rxd_include_node_flux3D.argtypes = [
    ctypes.c_int,
    _int_ptr,
    _int_ptr,
    _long_ptr,
    _double_ptr,
    ctypes.POINTER(ctypes.py_object),
]

_c_headers = """#include <math.h>
/*Some functions supported by numpy that aren't included in math.h
 * names and arguments match the wrappers used in rxdmath.py
 */
double factorial(const double);
double degrees(const double);
void radians(const double, double*);
double log1p(const double);
double vtrap(const double, const double);
"""


def _list_to_cint_array(data):
    if data is None or len(data) == 0:
        return None
    else:
        return (ctypes.c_int * len(data))(*tuple(data))


def _list_to_cdouble_array(data):
    if data is None or len(data) == 0:
        return None
    else:
        return (ctypes.c_double * len(data))(*tuple(data))


def _list_to_clong_array(data):
    if data is None or len(data) == 0:
        return None
    else:
        return (ctypes.c_long * len(data))(*tuple(data))


def _list_to_pyobject_array(data):
    if data is None or len(data) == 0:
        return None
    else:
        return (ctypes.py_object * len(data))(*tuple(data))


def byeworld():
    # do not call __del__ that rearrange memory for states
    species.Species.__del__ = lambda x: None
    species._ExtracellularSpecies.__del__ = lambda x: None
    species._IntracellularSpecies.__del__ = lambda x: None
    section1d.Section1D.__del__ = lambda x: None
    generalizedReaction.GeneralizedReaction.__del__ = lambda x: None

    # needed to prevent a seg-fault error at shutdown in at least some
    # combinations of NEURON and Python, which I think is due to objects
    # getting deleted out-of-order
    global _react_matrix_solver
    try:
        del _react_matrix_solver
    except NameError:
        #    # if it already didn't exist, that's fine
        pass
    _windows_remove_dlls()


atexit.register(byeworld)

_cvode_object = h.CVode()

last_diam_change_cnt = None
last_structure_change_cnt = None
last_nrn_legacy_units = h.nrnunit_use_legacy()


_all_reactions = []

nrn_tree_solve = nrn_dll_sym("nrn_tree_solve")
nrn_tree_solve.restype = None

_dptr = _double_ptr

_dimensions = {1: h.SectionList(), 3: h.SectionList()}
_dimensions_default = 1
_default_dx = 0.25
_default_method = "deterministic"

# CRxD
_diffusion_d = None
_diffusion_a = None
_diffusion_b = None
_diffusion_p = None
_diffusion_a_ptr, _diffusion_b_ptr, _diffusion_p_ptr = None, None, None


def _domain_lookup(sec, dim=None):
    for d, sl in _dimensions.items():
        if sec in sl:
            if dim is not None and d != dim:
                sl.remove(sec)
                return _domain_lookup(sec, dim)
            return d
    dimension = dim if dim else _dimensions_default
    _dimensions[dimension].append(sec)
    return dimension


def set_solve_type(domain=None, dimension=None, dx=None, nsubseg=None, method=None):
    """Specify the numerical discretization and solver options.

    domain -- a section or Python iterable of sections"""

    global _dimensions_default, _dimensions
    setting_default = False
    if domain is None:
        domain = h.allsec()
        setting_default = True
    elif isinstance(domain, nrn.Section):
        domain = [domain]

    # NOTE: These attributes are set on a per-nrn.Section basis; they cannot
    #       assume Section1D objects exist because they might be specified before
    #       those objects are created

    # domain is now always an iterable (or invalid)
    if method is not None:
        raise RxDException(
            "using set_solve_type to specify method is not yet implemented"
        )
    if dimension is not None:
        if dimension not in (1, 3):
            raise RxDException(
                "invalid option to set_solve_type: dimension must be 1 or 3"
            )
        if setting_default:
            _dimensions_default = dimension
        for sec in domain:
            _domain_lookup(sec, dimension)
    if dx is not None:
        raise RxDException("using set_solve_type to specify dx is not yet implemented")
    if nsubseg is not None:
        raise RxDException(
            "using set_solve_type to specify nsubseg is not yet implemented"
        )


def _unregister_reaction(r):
    global _all_reactions
    react = r() if isinstance(r, weakref.ref) else r
    with initializer._init_lock:
        _all_reactions = list(
            filter(lambda x: x() is not None and x() != react, _all_reactions)
        )


def _register_reaction(r):
    # TODO: should we search to make sure that (a weakref to) r hasn't already been added?
    global _all_reactions, _external_solver_initialized
    with initializer._init_lock:
        _all_reactions.append(_weakref_ref(r))
        _external_solver_initialized = False


def _after_advance():
    global last_diam_change_cnt
    last_diam_change_cnt = _diam_change_count.value


def re_init():
    """reinitializes all rxd concentrations to match HOC values, updates matrices"""

    global _external_solver_initialized
    h.define_shape()

    section1d._purge_cptrs()
    for sr in _species_get_all_species():
        s = sr()
        if s is not None:
            s._register_cptrs()

    for sr in _species_get_all_species():
        s = sr()
        if s is not None:
            s.re_init()

    # update matrix equations
    _setup_matrices()

    if _cvode_object.active():
        _cvode_object.re_init()

    _external_solver_initialized = False


def _setup_memb_currents():
    initializer._do_init()
    # setup membrane fluxes from our stuff
    # TODO: cache the memb_cur_ptrs, memb_cur_charges, memb_net_charges, memb_cur_mapped
    #       because won't change very often
    # need this; think it's because of initialization of mod files
    # setup for induced membrane currents
    cur_node_indices = []
    cur_map = {}
    curr_indices = []
    curr_scales = []
    curr_ptrs = []
    for sr in _species_get_all_species():
        s = sr()
        if s is not None:
            s._setup_currents(curr_indices, curr_scales, curr_ptrs, cur_map)
        num = len(curr_ptrs)
        if num:
            curr_ptr_vector = _h_ptrvector(num)
            curr_ptr_vector.ptr_update_callback(_donothing)
            for i, ptr in enumerate(curr_ptrs):
                curr_ptr_vector.pset(i, ptr)
                curr_ptr_storage_nrn = _h_vector(num)
        else:
            curr_ptr_vector = None
            curr_ptr_storage_nrn = None
    for rptr in _all_reactions:
        r = rptr()
        if r is not None:
            r._update_indices()
            r._setup_membrane_fluxes(cur_node_indices, cur_map)
    if not curr_indices:
        free_curr_ptrs()
        return
    rxd_setup_curr_ptrs(
        len(curr_indices),
        _list_to_cint_array(curr_indices),
        numpy.concatenate(curr_scales),
        _list_to_pyobject_array(curr_ptrs),
    )

    SPECIES_ABSENT = -1
    # TODO: change so that this is only called when there are in fact currents
    rxd_memb_scales = []
    memb_cur_ptrs = []
    memb_cur_mapped = []
    memb_cur_mapped_ecs = []
    memb_cur_ptrs = []
    for rptr in _all_reactions:
        r = rptr()
        if r and r._membrane_flux:
            r._do_memb_scales(cur_map)
            scales = r._memb_scales
            rxd_memb_scales.extend(scales)
            memb_cur_ptrs += r._cur_ptrs
            memb_cur_mapped += r._cur_mapped
            memb_cur_mapped_ecs += r._cur_mapped_ecs
    ecs_map = [
        SPECIES_ABSENT if i is None else i
        for i in list(
            itertools.chain.from_iterable(
                itertools.chain.from_iterable(memb_cur_mapped_ecs)
            )
        )
    ]
    ics_map = [
        SPECIES_ABSENT if i is None else i
        for i in list(
            itertools.chain.from_iterable(
                itertools.chain.from_iterable(memb_cur_mapped)
            )
        )
    ]
    if memb_cur_ptrs:
        cur_counts = [
            len(x) for x in memb_cur_mapped
        ]  # TODO: is len(x) the same for all x?
        num_fluxes = numpy.array(cur_counts).sum()
        num_currents = len(memb_cur_ptrs)
        memb_cur_ptrs = list(itertools.chain.from_iterable(memb_cur_ptrs))
        """print("num_currents",num_currents)
        print("num_fluxes",num_fluxes)
        print("num_nodes",curr_indices)
        print("num_species",cur_counts)
        print("cur_idxs",curr_indices)
        print("node_idxs",cur_node_indices)
        print("scales",rxd_memb_scales)
        print("ptrs",memb_cur_ptrs)
        print("mapped",ics_map,min(abs(numpy.array(ics_map))),max(ics_map))
        print("mapped_ecs",ecs_map,max(ecs_map))"""
        setup_currents(
            num_currents,
            num_fluxes,
            _list_to_cint_array(cur_counts),
            _list_to_cint_array(cur_node_indices),
            _list_to_cdouble_array(rxd_memb_scales),
            _list_to_pyobject_array(memb_cur_ptrs),
            _list_to_cint_array(ics_map),
            _list_to_cint_array(ecs_map),
        )


def _setup():
    from . import initializer

    if not initializer.is_initialized():
        initializer._do_init()
    # TODO: this is when I should resetup matrices (structure changed event)
    global _external_solver_initialized, last_diam_change_cnt, last_structure_change_cnt
    _external_solver_initialized = False

    # Using C-code for reactions
    options.use_reaction_contribution_to_jacobian = False
    with initializer._init_lock:
        _update_node_data()


def _find_librxdmath():
    import glob

    # cmake doesn't create x86_64 directory under install prefix
    base_path = os.path.join(h.neuronhome(), "..", "..", platform.machine())
    if not os.path.exists(base_path):
        base_path = os.path.join(h.neuronhome(), "..", "..")
    base_path = os.path.join(base_path, "lib", "librxdmath")
    success = False
    for extension in ["", ".dll", ".so", ".dylib"]:
        dll = base_path + extension
        try:
            success = os.path.exists(dll)
        except:
            pass
        if success:
            break
    if not success:
        if sys.platform.lower().startswith("win"):
            dll = os.path.join(h.neuronhome(), "bin", "librxdmath.dll")
            success = os.path.exists(dll)
        if not success:
            raise RxDException("unable to connect to the librxdmath library")
    return dll


def _c_compile(formula):
    filename = "rxddll" + str(uuid.uuid1())
    with open(filename + ".c", "w") as f:
        f.write(formula)
    math_library = "-lm"
    fpic = "-fPIC"
    try:
        gcc = os.environ["CC"]
    except:
        # when running on windows try and used the gcc included with NEURON
        if sys.platform.lower().startswith("win"):
            math_library = ""
            fpic = ""
            gcc = os.path.join(
                h.neuronhome(), "mingw", "mingw64", "bin", "x86_64-w64-mingw32-gcc.exe"
            )
            if not os.path.isfile(gcc):
                raise RxDException(
                    "unable to locate a C compiler. Please `set CC=<path to C compiler>`"
                )
        else:
            gcc = "gcc"
    # TODO: Check this works on non-Linux machines
    gcc_cmd = "%s -I%s -I%s " % (
        gcc,
        sysconfig.get_path("include"),
        os.path.join(h.neuronhome(), "..", "..", "include", "nrn"),
    )
    gcc_cmd += "-shared %s  %s.c %s " % (fpic, filename, _find_librxdmath())
    gcc_cmd += "-o %s.so %s" % (filename, math_library)
    if sys.platform.lower().startswith("win"):
        my_path = os.getenv("PATH")
        os.putenv(
            "PATH",
            my_path + ";" + os.path.join(h.neuronhome(), "mingw", "mingw64", "bin"),
        )
        os.system(gcc_cmd)
        os.putenv("PATH", my_path)
    else:
        os.system(gcc_cmd)
    # TODO: Find a better way of letting the system locate librxdmath.so.0
    rxdmath_dll = ctypes.cdll[_find_librxdmath()]
    dll = ctypes.cdll["%s.so" % os.path.abspath(filename)]
    reaction = dll.reaction
    reaction.argtypes = [
        ctypes.POINTER(ctypes.c_double),
        ctypes.POINTER(ctypes.c_double),
    ]
    reaction.restype = ctypes.c_double
    os.remove(filename + ".c")
    if sys.platform.lower().startswith("win"):
        # cannot remove dll that are in use
        _windows_dll.append(weakref.ref(dll))
        _windows_dll_files.append(filename + ".so")
    else:
        os.remove(filename + ".so")
    return reaction


_h_ptrvector = h.PtrVector
_h_vector = h.Vector

_structure_change_count = nrn_dll_sym("structure_change_cnt", _ctypes_c_int)
_diam_change_count = nrn_dll_sym("diam_change_cnt", _ctypes_c_int)


def _donothing():
    pass


def _setup_units(force=False):
    global last_nrn_legacy_units
    if initializer.is_initialized():
        if force or last_nrn_legacy_units != h.nrnunit_use_legacy():
            last_nrn_legacy_units = h.nrnunit_use_legacy()
            clear_rates()
            _setup_memb_currents()
            _compile_reactions()
            if _cvode_object.active():
                _cvode_object.re_init()


def _update_node_data(force=False, newspecies=False):
    global last_diam_change_cnt, last_structure_change_cnt
    if (
        last_diam_change_cnt != _diam_change_count.value
        or _structure_change_count.value != last_structure_change_cnt
        or force
    ):

        last_diam_change_cnt = _diam_change_count.value
        last_structure_change_cnt = _structure_change_count.value
        # if not species._has_3d:
        # TODO: merge this with the 3d/hybrid case?
        if initializer.is_initialized():
            nsegs_changed = 0
            for sr in _species_get_all_species():
                s = sr()
                if s is not None:
                    nsegs_changed += s._update_node_data()

                    # re-register ions for extracellular species in case new
                    # sections have been added within the ecs region
                    if hasattr(s, "_extracellular_instances"):
                        for ecs in s._extracellular_instances.values():
                            ecs._ion_register()

            if nsegs_changed or newspecies:
                section1d._purge_cptrs()
                for sr in _species_get_all_species():
                    s = sr()
                    if s is not None:
                        s._update_region_indices(True)
                        s._register_cptrs()
                # if species._has_1d and species._1d_submatrix_n():
                _setup_matrices()
                # TODO: separate compiling reactions -- so the indices can be updated without recompiling
                _include_flux(True)
                _setup_units(force=True)

            # end#if

            # _curr_scales = _numpy_array(_curr_scales)


def _matrix_to_rxd_sparse(m):
    """precondition: assumes m a numpy array"""
    nonzero_i, nonzero_j = list(zip(*list(m.keys())))
    nonzero_values = numpy.ascontiguousarray(list(m.values()), dtype=float)

    # number of rows
    n = m.shape[1]

    return (
        n,
        len(nonzero_i),
        numpy.ascontiguousarray(nonzero_i, dtype=numpy.int_),
        numpy.ascontiguousarray(nonzero_j, dtype=numpy.int_),
        nonzero_values,
    )


# TODO: make sure this does the right thing when the diffusion constant changes between two neighboring nodes
def _setup_matrices():

    with initializer._init_lock:

        # update _node_fluxes in C
        _include_flux()

        # TODO: this sometimes seems to get called twice. Figure out why and fix, if possible.

        n = len(_node_get_states())

        volumes = node._get_data()[0]
        zero_volume_indices = (numpy.where(volumes == 0)[0]).astype(numpy.int_)
        if species._has_1d:
            # TODO: initialization is slow. track down why

            _last_dt = None
            for sr in _species_get_all_species():
                s = sr()
                if s is not None:
                    s._assign_parents()

            # remove old linearmodeladdition
            _linmodadd_cur = None
            n = species._1d_submatrix_n()
            if n:
                # create sparse matrix for C in cy'+gy=b
                c_diagonal = numpy.zeros(n, dtype=ctypes.c_double)
                # most entries are 1 except those corresponding to the 0 and 1 ends

                # create the matrix G
                # if not species._has_3d:
                #    # if we have both, then put the 1D stuff into the matrix that already exists for 3D
                from collections import OrderedDict

                diffusion_matrix = [OrderedDict() for idx in range(n)]
                for sr in _species_get_all_species():
                    s = sr()
                    if s is not None:
                        s._setup_diffusion_matrix(diffusion_matrix)
                        s._setup_c_matrix(c_diagonal)
                        # print '_diffusion_matrix.shape = %r, n = %r, species._has_3d = %r' % (_diffusion_matrix.shape, n, species._has_3d)
                euler_matrix_i, euler_matrix_j, euler_matrix_nonzero = [], [], []
                for i in range(n):
                    mat_i = diffusion_matrix[i]
                    euler_matrix_i.extend(itertools.repeat(i, len(mat_i)))
                    euler_matrix_j.extend(mat_i.keys())
                    euler_matrix_nonzero.extend(mat_i.values())
                euler_matrix_nnonzero = len(euler_matrix_nonzero)
                assert (
                    len(euler_matrix_i)
                    == len(euler_matrix_j)
                    == len(euler_matrix_nonzero)
                )
                # modify C for cases where no diffusive coupling of 0, 1 ends
                # TODO: is there a better way to handle no diffusion?
                # for i in range(n):
                #    if not _diffusion_matrix[i, i]:
                #        _linmodadd_c[i, i] = 1

                # _cvode_object.re_init()

        # Hybrid logic
        if species._has_1d and species._has_3d:
            hybrid_neighbors = collections.defaultdict(lambda: [])
            hybrid_vols = collections.defaultdict(lambda: [])
            hybrid_diams = {}
            grid_id_dc = {}
            hybrid_index1d_grid_ids = {}
            grid_id_species = {}
            index1d_sec1d = {}
            hybrid_vols1d = {}
            dxs = set()
            for sr in _species_get_all_species():
                s = sr()
                if s is not None:
                    if s._intracellular_instances:
                        # have both 1D and 3D, so find the neighbors
                        # for each of the 3D sections, find the parent sections
                        for r in s._regions:
                            if r in s._intracellular_instances:
                                grid_id = s._intracellular_instances[r]._grid_id
                                grid_id_species.setdefault(
                                    grid_id, s._intracellular_instances[r]
                                )
                                grid_id_dc[grid_id] = s.d
                                dxs.add(r._dx)
                                for sec in r._secs3d:
                                    parent_seg = sec.trueparentseg()
                                    parent_sec = (
                                        None if not parent_seg else parent_seg.sec
                                    )
                                    # are any of these a match with a 1d section?
                                    if s._has_region_section(r, parent_sec):
                                        # this section has a 1d section that is a parent
                                        (
                                            index1d,
                                            indices3d,
                                            vols1d,
                                            vols3d,
                                        ) = _get_node_indices(
                                            s,
                                            r,
                                            sec,
                                            sec.orientation(),
                                            parent_sec,
                                            h.parent_connection(sec=sec),
                                        )
                                        hybrid_neighbors[index1d] += indices3d
                                        hybrid_vols[index1d] += vols3d
                                        hybrid_diams[index1d] = parent_sec(
                                            h.parent_connection(sec=sec)
                                        ).diam
                                        hybrid_index1d_grid_ids[index1d] = grid_id
                                        index1d_sec1d[index1d] = parent_sec
                                        hybrid_vols1d[index1d] = vols1d

                                    for sec1d in r._secs1d:
                                        parent_1d_seg = sec1d.trueparentseg()
                                        parent_1d = (
                                            None
                                            if not parent_1d_seg
                                            else parent_1d_seg.sec
                                        )
                                        if parent_1d == sec:
                                            # it is the parent of a 1d section
                                            (
                                                index1d,
                                                indices3d,
                                                vols1d,
                                                vols3d,
                                            ) = _get_node_indices(
                                                s,
                                                r,
                                                sec,
                                                parent_1d_seg.x,
                                                sec1d,
                                                sec1d.orientation(),
                                            )
                                            hybrid_neighbors[index1d] += indices3d
                                            hybrid_vols[index1d] += vols3d
                                            hybrid_diams[index1d] = sec1d(
                                                h.section_orientation(sec=sec1d)
                                            ).diam
                                            hybrid_index1d_grid_ids[index1d] = grid_id
                                            index1d_sec1d[index1d] = sec1d
                                            hybrid_vols1d[index1d] = vols1d

            if len(dxs) > 1:
                raise RxDException("currently require a unique value for dx")
            dx = dxs.pop()
            rates = []
            volumes3d = []
            volumes1d = []
            grids_dx = []
            hybrid_indices1d = []
            hybrid_indices3d = []
            num_3d_indices_per_1d_seg = []

            num_1d_indices_per_grid = []
            num_3d_indices_per_grid = []

            grid_id_indices1d = collections.defaultdict(lambda: [])
            for index1d in hybrid_neighbors:
                grid_id = hybrid_index1d_grid_ids[index1d]
                grid_id_indices1d[grid_id].append(index1d)
            hybrid_grid_ids = sorted(grid_id_indices1d.keys())
            for grid_id in hybrid_grid_ids:
                sp = grid_id_species[grid_id]
                # TODO: use 3D anisotropic diffusion coefficients
                dc = grid_id_dc[grid_id]
                grids_dx.append(sp._dx**3)
                num_1d_indices_per_grid.append(len(grid_id_indices1d[grid_id]))
                grid_3d_indices_cnt = 0
                for index1d in grid_id_indices1d[grid_id]:
                    neighbors3d = []
                    vols3d = []
                    for neigh, vol in zip(
                        hybrid_neighbors[index1d], hybrid_vols[index1d]
                    ):
                        if neigh not in neighbors3d:
                            neighbors3d.append(neigh)
                            vols3d.append(vol)
                    if len(neighbors3d) < 1:
                        raise RxDException(
                            "No 3D neighbors detected for 1D segment. Try perturbing dx"
                        )
                    sec1d = index1d_sec1d[index1d]
                    seg_length1d = sec1d.L / sec1d.nseg
                    if neighbors3d:
                        hybrid_indices1d.append(index1d)
                        cnt_neighbors_3d = len(neighbors3d)
                        num_3d_indices_per_1d_seg.append(cnt_neighbors_3d)
                        grid_3d_indices_cnt += cnt_neighbors_3d
                        area = numpy.pi * 0.25 * hybrid_diams[index1d] ** 2
                        areaT = sum([v ** (2.0 / 3.0) for v in vols3d])
                        volumes1d.append(hybrid_vols1d[index1d])
                        for i, vol in zip(neighbors3d, vols3d):
                            sp._region._vol[i] = vol
                            ratio = vol ** (2.0 / 3.0) / areaT
                            rate = ratio * dc * area / (vol * (dx + seg_length1d) / 2)
                            rates.append(rate)
                            volumes3d.append(vol)
                            hybrid_indices3d.append(i)

                num_3d_indices_per_grid.append(grid_3d_indices_cnt)

            num_1d_indices_per_grid = numpy.asarray(
                num_1d_indices_per_grid, dtype=numpy.int64
            )
            num_3d_indices_per_grid = numpy.asarray(
                num_3d_indices_per_grid, dtype=numpy.int64
            )

            hybrid_indices1d = numpy.asarray(hybrid_indices1d, dtype=numpy.int64)
            num_3d_indices_per_1d_seg = numpy.asarray(
                num_3d_indices_per_1d_seg, dtype=numpy.int64
            )
            hybrid_grid_ids = numpy.asarray(hybrid_grid_ids, dtype=numpy.int64)

            hybrid_indices3d = numpy.asarray(hybrid_indices3d, dtype=numpy.int64)
            rates = numpy.asarray(rates, dtype=float)
            volumes1d = numpy.asarray(volumes1d, dtype=float)
            volumes3d = numpy.asarray(volumes3d, dtype=float)
            dxs = numpy.asarray(grids_dx, dtype=float)
            set_hybrid_data(
                num_1d_indices_per_grid,
                num_3d_indices_per_grid,
                hybrid_indices1d,
                hybrid_indices3d,
                num_3d_indices_per_1d_seg,
                hybrid_grid_ids,
                rates,
                volumes1d,
                volumes3d,
                dxs,
            )

        # TODO: Replace this this to handle 1d/3d hybrid models
        """
        if species._has_1d and species._has_3d:
            # TODO: add connections to matrix; for now: find them
            hybrid_neighbors = collections.defaultdict(lambda: [])
            hybrid_diams = {}
            dxs = set()
            for sr in list(_species_get_all_species().values()):
                s = sr()
                if s is not None:
                    if s._nodes and s._secs:
                        # have both 1D and 3D, so find the neighbors
                        # for each of the 3D sections, find the parent sections
                        for r in s._regions:
                            dxs.add(r._dx)
                            for sec in r._secs3d:
                                parent_seg = sec.trueparentseg()
                                parent_sec = None if not parent_seg else parent_seg.sec
                                # are any of these a match with a 1d section?
                                if s._has_region_section(r, parent_sec):
                                    # this section has a 1d section that is a parent
                                    index1d, indices3d = _get_node_indices(s, r, sec, h.section_orientation(sec=sec), parent_sec, h.parent_connection(sec=sec))
                                    hybrid_neighbors[index1d] += indices3d
                                    hybrid_diams[index1d] = parent_seg.diam
                                else:
                                    for sec1d in r._secs1d:
                                        parent_1d_seg = sec1d.trueparentseg()
                                        parent_1d = None if not parent_seg else parent_seg.sec
                                        if parent_1d == sec:
                                            # it is the parent of a 1d section
                                            index1d, indices3d = _get_node_indices(s, r, sec, h.parent_connection(sec=sec1d), sec1d, sec1d.orientation())
                                            hybrid_neighbors[index1d] += indices3d
                                            hybrid_diams[index1d] = parent_1d_seg.diam
                                            break
                                        elif parent_1d == parent_sec:
                                            # it connects to the parent of a 1d section
                                            index1d, indices3d = _get_node_indices(s, r, sec, h.section_orientation(sec=sec), sec1d, sec1d.orientation())
                                            hybrid_neighbors[index1d] += indices3d
                                            hybrid_diams[index1d] = parent_1d_seg.diam
                                            break
            if len(dxs) > 1:
                raise RxDException('currently require a unique value for dx')
            dx = dxs.pop()
            diffs = node._diffs
            n = len(_node_get_states())
            # TODO: validate that we're doing the right thing at boundaries
            for index1d in list(hybrid_neighbors.keys()):
                neighbors3d = set(hybrid_neighbors[index1d])
                # NOTE: splitting the connection area equally across all the connecting nodes
                area = (numpy.pi * 0.25 * hybrid_diams[index1d] ** 2) / len(neighbors3d)
                for i in neighbors3d:
                    d = diffs[i]
                    vol = node._volumes[i]
                    rate = d * area / (vol * dx / 2.)
                    # make the connections on the 3d side
                    _euler_matrix[i, i] -= rate
                    _euler_matrix[i, index1d] += rate
                    # make the connections on the 1d side (scale by vol because conserving mass not volume)
                    _euler_matrix[index1d, index1d] -= rate * vol
                    _euler_matrix[index1d, i] += rate * vol
                #print 'index1d row sum:', sum(_euler_matrix[index1d, j] for j in xrange(n))
                #print 'index1d col sum:', sum(_euler_matrix[j, index1d] for j in xrange(n))
        """
        # CRxD
        setup_solver(
            _node_get_states(),
            len(_node_get_states()),
            zero_volume_indices,
            len(zero_volume_indices),
        )
        if species._has_1d and n and euler_matrix_nnonzero > 0:
            section1d._transfer_to_legacy()
            set_euler_matrix(
                n,
                euler_matrix_nnonzero,
                _list_to_clong_array(euler_matrix_i),
                _list_to_clong_array(euler_matrix_j),
                _list_to_cdouble_array(euler_matrix_nonzero),
                c_diagonal,
            )
        else:
            rxd_set_no_diffusion()

        if section1d._all_cindices is not None and len(section1d._all_cindices) > 0:
            rxd_setup_conc_ptrs(
                len(section1d._all_cindices),
                _list_to_cint_array(section1d._all_cindices),
                _list_to_pyobject_array(section1d._all_cptrs),
            )
        else:
            free_conc_ptrs()

    # we do this last because of performance issues with changing sparsity of csr matrices
    """
    if _diffusion_matrix is not None:
        _diffusion_matrix = _diffusion_matrix.tocsr()
    if _euler_matrix is not None:
        _euler_matrix = _euler_matrix.tocsr()

    if species._has_1d:
        if species._has_3d:
            _diffusion_matrix = -_euler_matrix
        n = species._1d_submatrix_n()
        if n:
            matrix = _diffusion_matrix[_zero_volume_indices].tocsr()
            indptr = matrix.indptr
            matrixdata = matrix.data
            count = len(_zero_volume_indices)
            for row, i in enumerate(_zero_volume_indices):
                d = _diffusion_matrix[i, i]
                if d:
                    matrixdata[indptr[row] : indptr[row + 1]] /= -d
                    matrix[row, i] = 0
                else:
                    matrixdata[indptr[row] : indptr[row + 1]] = 0
            global _mat_for_zero_volume_nodes
            _mat_for_zero_volume_nodes = matrix
            # TODO: _mat_for_zero_volume_nodes is used for CVode.
            #       Figure out if/how it has to be changed for hybrid 1D/3D sims (probably just augment with identity? or change how its used to avoid multiplying by I)
    
    """

    """
    if pt1 in indices:
        ileft = indices[pt1]
        dleft = (d + diffs[ileft]) * 0.5
        left = dleft * areal / (vol * dx)
        euler_matrix[index, ileft] += left
        euler_matrix[index, index] -= left
    if pt2 in indices:
        iright = indices[pt2]
        dright = (d + diffs[iright]) * 0.5
        right = dright * arear / (vol * dx)
        euler_matrix[index, iright] += right
        euler_matrix[index, index] -= right
"""


def _get_node_indices(species, region, sec3d, x3d, sec1d, x1d):
    # Recalculate the volumes
    xlo, xhi = region._mesh_grid["xlo"], region._mesh_grid["xhi"]
    ylo, yhi = region._mesh_grid["ylo"], region._mesh_grid["yhi"]
    zlo, zhi = region._mesh_grid["zlo"], region._mesh_grid["zhi"]
    from . import geometry3d

    p3d = int((sec3d.n3d() - 1) * x3d)
    p1d = int((sec1d.n3d() - 1) * x1d)
    pt3d = [p3d, p3d + 1] if p3d == 0 else [p3d - 1, p3d]
    pt1d = [p1d, p1d + 1] if p1d == 0 else [p1d - 1, p1d]

    inter, surf, mesh = geometry3d.voxelize2(
        [sec1d, sec3d],
        region._dx,
        mesh_grid=region._mesh_grid,
        relevant_pts=[pt1d, pt3d],
    )

    # TODO: remove need for this assumption
    assert x1d in (0, 1)
    disc_indices = region._indices_from_sec_x(sec3d, x3d)
    # print '%r(%g) connects to the 1d section %r(%g)' % (sec3d, x3d, sec1d, x1d)
    # print 'disc indices: %r' % disc_indices
    indices3d = []
    vols3d = []
    for point in disc_indices:
        if (
            point in _point_indices[region]
            and _point_indices[region][point] not in indices3d
        ):
            indices3d.append(_point_indices[region][point])
            vols3d.append(surf[point][0] if point in surf else region.dx ** 3)
            # print 'found node %d with coordinates (%g, %g, %g)' % (node._index, node.x3d, node.y3d, node.z3d)
    # discard duplicates...
    # TODO: really, need to figure out all the 3d nodes connecting to a given 1d endpoint, then unique that
    # print '3d matrix indices: %r' % indices3d
    # TODO: remove the need for this assertion
    if x1d == sec1d.orientation():
        # TODO: make this whole thing more efficient
        # the parent node is the nonzero index on the first row before the diagonal
        # first_row = min([node._index for node in species.nodes(region)(sec1d)])
        index_1d, vol1d = min(
            [(node._index, node.volume) for node in species.nodes(region)(sec1d)],
            key=lambda x: x[0],
        )
        """for j in range(first_row):
            if _euler_matrix[first_row, j] != 0:
                index_1d = j
                break
        else:
            raise RxDException('should never get here; could not find parent')"""
    elif x1d == 1 - sec1d.orientation():
        # the ending zero-volume node is the one after the last node
        # TODO: make this more efficient
        index_1d, vol1d = max(
            [(node._index, node.volume) for node in species.nodes(region)(sec1d)],
            key=lambda x: x[0],
        )
        index_1d + 1
    else:
        raise RxDException(
            "should never get here; _get_node_indices apparently only partly converted to allow connecting to 1d in middle"
        )
    return index_1d, indices3d, vol1d, vols3d


def _compile_reactions():
    # clear all previous reactions (intracellular & extracellular) and the
    # supporting indexes
    # _windows_remove_dlls()

    regions_inv = dict()  # regions -> reactions that occur there
    species_by_region = dict()
    all_species_involed = set()
    location_count = 0

    ecs_regions_inv = dict()
    ecs_species_by_region = dict()
    ecs_all_species_involed = set()
    ecs_mc_species_involved = set()
    from . import rate, multiCompartmentReaction

    # Find sets of sections that contain the same regions
    from .region import _c_region

    matched_regions = (
        []
    )  # the different combinations of regions that arise in different sections
    rxd_sec_lookup = section1d._SectionLookup()
    for nrnsec in rxd_sec_lookup:
        set_of_regions = set()  # a set of the regions that occur in a given section
        for sec in rxd_sec_lookup[nrnsec]:
            if sec:
                set_of_regions.add(sec._region)
        if set_of_regions not in matched_regions:
            matched_regions.append(set_of_regions)
    region._c_region_lookup = dict()

    # create a c_region instance for each of the unique sets of regions
    c_region_list = []
    for sets in matched_regions:
        c_region_list.append(_c_region(sets))

    for rptr in _all_reactions:
        r = rptr()
        if not r:
            continue

        # Find all the species involved
        if isinstance(r, rate.Rate):
            if not r._species():
                continue
            sptrs = set([r._species])
        else:
            sptrs = set(r._dests + r._sources)

        if hasattr(r, "_involved_species") and r._involved_species:
            sptrs = sptrs.union(set(r._involved_species))
        if hasattr(r, "_involved_species_ecs") and r._involved_species_ecs:
            sptrs = sptrs.union(set(r._involved_species_ecs))
        # Find all the regions involved
        if isinstance(r, multiCompartmentReaction.MultiCompartmentReaction):
            if not hasattr(r._mult, "flatten"):
                r._update_indices()
            react_regions = [
                s()._extracellular()._region
                for s in r._sources + r._dests
                if isinstance(s(), species.SpeciesOnExtracellular)
            ] + [
                s()._region()
                for s in r._sources + r._dests
                if not isinstance(s(), species.SpeciesOnExtracellular)
            ]
            react_regions += [
                sptr()._region()
                for sptr in sptrs
                if isinstance(sptr(), species.SpeciesOnRegion)
            ]
            react_regions += [r._regions[0]]
            react_regions = list(set(react_regions))

        # if regions are specified - use those
        elif hasattr(r, "_active_regions"):
            react_regions = r._active_regions
        # Otherwise use all the regions where the species are
        else:
            react_regions = set()
            nsp = 0
            for sp in sptrs:
                s = sp()
                nsp += 1
                if isinstance(s, species.SpeciesOnRegion):
                    react_regions.add(s._region())
                elif isinstance(s, species.SpeciesOnExtracellular):
                    react_regions.add(s._extracellular()._region)
                elif isinstance(s, species._ExtracellularSpecies):
                    react_regions.add(s._region)
                elif None not in s._regions:
                    [
                        react_regions.add(reg)
                        for reg in s._regions + s._extracellular_regions
                    ]
            react_regions = list(react_regions)
            # Only regions where ALL the species are present -- unless it is a membrane
            # from collections import Counter
            # from . import geometry as geo
            # react_regions = [reg for reg, count in Counter(react_regions).iteritems() if count == nsp or isinstance(reg.geometry,geo.ScalableBorder)]
        # Any intracellular regions
        if not all([isinstance(x, region.Extracellular) for x in react_regions]):
            species_involved = []
            for sp in sptrs:
                s = sp()
                if not isinstance(s, species.SpeciesOnExtracellular) and not isinstance(
                    s, species._ExtracellularSpecies
                ):
                    all_species_involed.add(s)
                    species_involved.append(s)
            for reg in react_regions:
                if isinstance(reg, region.Extracellular):
                    continue
                if reg in regions_inv:
                    regions_inv[reg].append(rptr)
                else:
                    regions_inv[reg] = [rptr]
                if reg in species_by_region:
                    species_by_region[reg] = species_by_region[reg].union(
                        species_involved
                    )
                else:
                    species_by_region[reg] = set(species_involved)
                    for sec in reg._secs:
                        location_count += sec.nseg
        # Any extracellular regions
        if any([isinstance(x, region.Extracellular) for x in react_regions]):
            # MultiCompartment - so can have both extracellular and intracellular regions
            if isinstance(r, multiCompartmentReaction.MultiCompartmentReaction):
                for sp in sptrs:
                    s = sp()
                    if isinstance(s, species.SpeciesOnExtracellular):
                        ecs_mc_species_involved.add(s)
                    if isinstance(s, species.Species) and s._extracellular_instances:
                        for ecs in s._extracellular_instances.keys():
                            ecs_mc_species_involved.add(s[ecs])
                for reg in react_regions:
                    if reg in list(ecs_species_by_region.keys()):
                        ecs_species_by_region[reg] = ecs_species_by_region[reg].union(
                            ecs_mc_species_involved
                        )
                    else:
                        ecs_species_by_region[reg] = set(ecs_mc_species_involved)
            # Otherwise - reaction can only have extracellular regions
            else:
                ecs_species_involved = []
                for sp in sptrs:
                    s = sp()
                    ecs_all_species_involed.add(s)
                    ecs_species_involved.append(s)
                for reg in react_regions:
                    if not isinstance(reg, region.Extracellular):
                        continue
                    if reg in ecs_regions_inv:
                        ecs_regions_inv[reg].append(rptr)
                    else:
                        ecs_regions_inv[reg] = [rptr]
                    if reg in ecs_species_by_region:
                        ecs_species_by_region[reg] = ecs_species_by_region[reg].union(
                            ecs_species_involved
                        )
                    else:
                        ecs_species_by_region[reg] = set(ecs_species_involved)
    # Create lists of indexes for intracellular reactions and rates
    # a table for location,species -> state index
    regions_inv_1d = [reg for reg in regions_inv if any(reg._secs1d)]
    regions_inv_1d.sort(key=lambda r: r._id)
    all_regions_inv_3d = [reg for reg in regions_inv if any(reg._secs3d)]
    # remove extra regions from multicompartment reactions. We only want the membrane
    regions_inv_3d = set()
    for reg in all_regions_inv_3d:
        for rptr in regions_inv[reg]:
            r = rptr()
            if isinstance(r, multiCompartmentReaction.MultiCompartmentReaction):
                regions_inv_3d.add(r._regions[0])
            else:
                regions_inv_3d.add(reg)
    regions_inv_3d = list(regions_inv_3d)

    for reg in regions_inv_1d:
        rptr = weakref.ref(reg)
        if rptr in region._c_region_lookup:
            for c_region in region._c_region_lookup[rptr]:
                for react in regions_inv[reg]:
                    c_region.add_reaction(react, rptr)
                    c_region.add_species(species_by_region[reg])
                    if reg in ecs_species_by_region:
                        c_region.add_ecs_species(ecs_species_by_region[reg])

    # now setup the reactions
    # if there are no reactions
    if location_count == 0 and len(ecs_regions_inv) == 0:
        return None

    def localize_index(creg, rate):
        rate_str = re.sub(
            r"species\[(\d+)\]\[(\d+)\]",
            lambda m: "species[%i][%i]"
            % (
                creg._species_ids.get(int(m.groups()[0])),
                creg._region_ids.get(int(m.groups()[1])),
            ),
            rate,
        )
        rate_str = re.sub(
            r"params\[(\d+)\]\[(\d+)\]",
            lambda m: "params[%i][%i]"
            % (
                creg._params_ids.get(int(m.groups()[0])),
                creg._region_ids.get(int(m.groups()[1])),
            ),
            rate_str,
        )
        rate_str = re.sub(
            r"species_3d\[(\d+)\]",
            lambda m: "species_3d[%i]" % creg._ecs_species_ids.get(int(m.groups()[0])),
            rate_str,
        )
        rate_str = re.sub(
            r"params_3d\[(\d+)\]",
            lambda m: "params_3d[%i]" % creg._ecs_params_ids.get(int(m.groups()[0])),
            rate_str,
        )
        return rate_str

    # Setup intracellular and multicompartment reactions
    if location_count > 0:
        from . import rate, multiCompartmentReaction, Parameter

        for creg in c_region_list:
            if not creg._react_regions:
                continue
            creg._initalize()
            mc_mult_count = 0
            mc_mult_list = []
            species_ids_used = numpy.zeros((creg.num_species, creg.num_regions), bool)
            flux_ids_used = numpy.zeros((creg.num_species, creg.num_regions), bool)
            ecs_species_ids_used = numpy.zeros((creg.num_ecs_species), bool)
            fxn_string = _c_headers
            fxn_string += "void reaction(double** species, double** params, double** rhs, double* mult, double* species_3d, double* params_3d, double* rhs_3d, double** flux, double v)\n{"
            # declare the "rate" variable if any reactions (non-rates)
            for rprt in creg._react_regions:
                if not isinstance(rprt(), rate.Rate):
                    fxn_string += "\n\tdouble rate;"
                    break
            for rptr in _all_reactions:
                if rptr not in creg._react_regions:
                    continue
                r = rptr()
                if isinstance(r, rate.Rate):
                    s = r._species()
                    if s._id in creg._params_ids:
                        warn(
                            "Parameters values are fixed, %r will not change the value of %r"
                            % (r, s)
                        )
                        continue
                    species_id = creg._species_ids[s._id]
                    for reg in creg._react_regions[rptr]:
                        if reg() in r._rate:
                            try:
                                region_id = creg._region_ids[reg()._id]
                                rate_str = localize_index(creg, r._rate[reg()][0])
                            except KeyError:
                                warn(
                                    "Species not on the region specified, %r will be ignored.\n"
                                    % r
                                )
                                continue
                            operator = (
                                "+=" if species_ids_used[species_id][region_id] else "="
                            )
                            fxn_string += "\n\trhs[%d][%d] %s %s;" % (
                                species_id,
                                region_id,
                                operator,
                                rate_str,
                            )
                            species_ids_used[species_id][region_id] = True
                elif isinstance(r, multiCompartmentReaction.MultiCompartmentReaction):
                    # Lookup the region_id for the reaction
                    try:
                        for reg in r._rate:
                            rate_str = localize_index(creg, r._rate[reg][0])
                            fxn_string += "\n\trate = %s;" % rate_str
                            break
                    except KeyError:
                        warn(
                            "Species not on the region specified, %r will be ignored.\n"
                            % r
                        )
                        continue

                    for i, sptr in enumerate(r._sources + r._dests):
                        s = sptr()
                        if isinstance(s, species.SpeciesOnExtracellular):
                            if not isinstance(s, species.ParameterOnExtracellular):
                                species_id = creg._ecs_species_ids[
                                    s._extracellular()._grid_id
                                ]
                                operator = (
                                    "+=" if ecs_species_ids_used[species_id] else "="
                                )
                                fxn_string += "\n\trhs_3d[%d] %s mult[%d] * rate;" % (
                                    species_id,
                                    operator,
                                    mc_mult_count,
                                )
                                ecs_species_ids_used[species_id] = True
                        elif not isinstance(s, species.Parameter) and not isinstance(
                            s, species.ParameterOnRegion
                        ):
                            species_id = creg._species_ids[s._id]
                            region_id = creg._region_ids[s._region()._id]
                            operator = (
                                "+=" if species_ids_used[species_id][region_id] else "="
                            )
                            fxn_string += "\n\trhs[%d][%d] %s mult[%d] * rate;" % (
                                species_id,
                                region_id,
                                operator,
                                mc_mult_count,
                            )
                            species_ids_used[species_id][region_id] = True
                            if r._membrane_flux:
                                operator = (
                                    "+="
                                    if flux_ids_used[species_id][region_id]
                                    else "="
                                )
                                fxn_string += (
                                    "\n\tif(flux) flux[%d][%d] %s %1.1f * rate;"
                                    % (
                                        species_id,
                                        region_id,
                                        operator,
                                        r._cur_charges[i],
                                    )
                                )
                                flux_ids_used[species_id][region_id] = True
                            # TODO: Fix problem if the whole region isn't part of the same aggregate c_region
                        mc_mult_count += 1
                    mc_mult_list.extend(r._mult.flatten())
                else:
                    for reg in creg._react_regions[rptr]:
                        try:
                            region_id = creg._region_ids[reg()._id]
                            rate_str = localize_index(creg, r._rate[reg()][0])
                        except KeyError:
                            warn(
                                "Species not on the region specified, %r will be ignored.\n"
                                % r
                            )
                            continue
                        fxn_string += "\n\trate = %s;" % rate_str
                        summed_mults = collections.defaultdict(lambda: 0)
                        for (mult, sp) in zip(r._mult, r._sources + r._dests):
                            summed_mults[creg._species_ids.get(sp()._id)] += mult
                        for idx in sorted([k for k in summed_mults if k is not None]):
                            operator = "+=" if species_ids_used[idx][region_id] else "="
                            species_ids_used[idx][region_id] = True
                            fxn_string += "\n\trhs[%d][%d] %s (%g) * rate;" % (
                                idx,
                                region_id,
                                operator,
                                summed_mults[idx],
                            )
            fxn_string += "\n}\n"
            register_rate(
                creg.num_species,
                creg.num_params,
                creg.num_regions,
                creg.num_segments,
                creg.get_state_index(),
                creg.num_ecs_species,
                creg.num_ecs_params,
                creg.get_ecs_species_ids(),
                creg.get_ecs_index(),
                mc_mult_count,
                numpy.array(mc_mult_list, dtype=ctypes.c_double),
                _list_to_pyobject_array(creg._vptrs),
                _c_compile(fxn_string),
            )

    # Setup intracellular 3D reactions
    molecules_per_mM_um3 = constants.molecules_per_mM_um3()
    if regions_inv_3d:
        for reg in regions_inv_3d:
            ics_grid_ids = []
            all_ics_gids = set()
            ics_param_gids = set()
            fxn_string = _c_headers
            fxn_string += "void reaction(double* species_3d, double* params_3d, double*rhs, double* mc3d_mults)\n{"
            for rptr in [r for rlist in list(regions_inv.values()) for r in rlist]:
                if not isinstance(rptr(), rate.Rate):
                    fxn_string += "\n\tdouble rate;\n"
                    break
            # if any rates on this region have SpeciesOnRegion, add their grid_ids
            # do this in loop above if it is correct
            for rptr in [r for rlist in list(regions_inv.values()) for r in rlist]:
                r = rptr()
                if isinstance(r, rate.Rate):
                    if reg in r._regions:
                        for spec_involved in r._involved_species:
                            # probably should do parameters/states here as well
                            if isinstance(spec_involved(), species.SpeciesOnRegion):
                                all_ics_gids.add(
                                    spec_involved()
                                    ._species()
                                    ._intracellular_instances[spec_involved()._region()]
                                    ._grid_id
                                )
                elif isinstance(r, multiCompartmentReaction.MultiCompartmentReaction):
                    if reg in r._rate:
                        for spec_involved in (
                            r._involved_species + r._sources + r._dests
                        ):
                            all_ics_gids.add(
                                spec_involved()
                                ._species()
                                ._intracellular_instances[spec_involved()._region()]
                                ._grid_id
                            )

            for s in species_by_region[reg]:
                spe = s._species() if isinstance(s, species.SpeciesOnRegion) else s
                if (
                    hasattr(spe, "_intracellular_instances")
                    and spe._intracellular_instances
                    and reg in spe._intracellular_instances
                ):
                    if isinstance(s, species.Parameter) or isinstance(
                        s, species.ParameterOnRegion
                    ):
                        sp = spe._intracellular_instances[reg]
                        ics_param_gids.add(sp._grid_id)
                    else:
                        ###TODO is this correct? are there any other cases I should worry about? Do I always make a species the intracellular instance for the region we are looping through?
                        sp = spe._intracellular_instances[reg]
                        all_ics_gids.add(sp._grid_id)
            all_ics_gids = list(all_ics_gids)
            ics_param_gids = list(ics_param_gids)
            if any(
                [
                    isinstance(
                        rptr(), multiCompartmentReaction.MultiCompartmentReaction
                    )
                    for rptr in regions_inv[reg]
                ]
            ):
                # the elements in each list contain the indices into the states vector for the intracellular instance that need to be updated
                mc3d_region_size = len(reg._xs)
                mc3d_indices_start = [
                    species._defined_species_by_gid[index]._mc3d_indices_start(reg)
                    for index in all_ics_gids + ics_param_gids
                ]
            else:
                mc3d_region_size = 0
                mc3d_indices_start = [
                    0 for i in range(len(all_ics_gids + ics_param_gids))
                ]
            mults = [[] for i in range(len(all_ics_gids + ics_param_gids))]
            for rptr in regions_inv[reg]:
                r = rptr()
                if reg not in r._rate:
                    continue
                rate_str = re.sub(
                    r"species_3d\[(\d+)\]",
                    lambda m: "species_3d[%i]"
                    % [
                        pid
                        for pid, gid in enumerate(all_ics_gids)
                        if gid == int(m.groups()[0])
                    ][0],
                    r._rate[reg][-1],
                )
                rate_str = re.sub(
                    r"params_3d\[(\d+)\]",
                    lambda m: "params_3d[%i]"
                    % [
                        pid
                        for pid, gid in enumerate(ics_param_gids)
                        if gid == int(m.groups()[0])
                    ][0],
                    rate_str,
                )
                if isinstance(r, rate.Rate):
                    s = r._species()
                    # Get underlying rxd._IntracellularSpecies for the grid_id
                    if isinstance(s, species.Parameter) or isinstance(
                        s, species.ParameterOnRegion
                    ):
                        continue
                    elif isinstance(s, species.Species):
                        s = s._intracellular_instances[reg]
                    elif isinstance(s, species.SpeciesOnRegion):
                        s = s._species()._intracellular_instances[s._region()]
                    if s._grid_id in ics_grid_ids:
                        operator = "+="
                    else:
                        operator = "="
                        ics_grid_ids.append(s._grid_id)
                    pid = [
                        pid for pid, gid in enumerate(all_ics_gids) if gid == s._grid_id
                    ][0]
                    fxn_string += "\n\trhs[%d] %s %s;" % (pid, operator, rate_str)
                elif isinstance(r, multiCompartmentReaction.MultiCompartmentReaction):
                    if reg in r._regions:
                        from . import geometry

                        fxn_string += "\n\trate = " + rate_str + ";"
                        for sptr in r._sources:
                            s = sptr()
                            if not isinstance(s, species.Parameter) and not isinstance(
                                s, species.ParameterOnRegion
                            ):
                                s3d = s.instance3d
                                if s3d._grid_id in ics_grid_ids:
                                    operator = "+="
                                else:
                                    operator = "="
                                    ics_grid_ids.append(s3d._grid_id)
                                    # Find mult for this grid
                                    sas = reg._vol
                                    s3d_reg = s3d._region
                                    for segidx, seg in enumerate(s3d_reg._segs3d()):
                                        # Change this to be by volume
                                        # membrane area / compartment volume / molecules_per_mM_um3
                                        mults[s3d._grid_id] += [
                                            sas[index]
                                            / (s3d_reg._vol[index])
                                            / molecules_per_mM_um3
                                            for index in s3d_reg._nodes_by_seg[segidx]
                                        ]

                                pid = [
                                    pid
                                    for pid, gid in enumerate(all_ics_gids)
                                    if gid == s3d._grid_id
                                ][0]
                                fxn_string += (
                                    "\n\trhs[%d] %s -mc3d_mults[%d] * rate;"
                                    % (pid, operator, pid)
                                )
                        for sptr in r._dests:
                            s = sptr()
                            if not isinstance(s, species.Parameter) and not isinstance(
                                s, species.ParameterOnRegion
                            ):
                                s3d = s.instance3d
                                if s3d._grid_id in ics_grid_ids:
                                    operator = "+="
                                else:
                                    operator = "="
                                    ics_grid_ids.append(s3d._grid_id)
                                    # Find mult for this grid
                                    sas = reg._vol
                                    s3d_reg = s3d._region
                                    for segidx, seg in enumerate(s3d_reg._segs3d()):
                                        # Change this to be by volume
                                        mults[s3d._grid_id] += [
                                            sas[index]
                                            / (s3d_reg._vol[index])
                                            / molecules_per_mM_um3
                                            for index in s3d_reg._nodes_by_seg[segidx]
                                        ]
                                pid = [
                                    pid
                                    for pid, gid in enumerate(all_ics_gids)
                                    if gid == s3d._grid_id
                                ][0]
                                fxn_string += (
                                    "\n\trhs[%d] %s mc3d_mults[%d] * rate;"
                                    % (pid, operator, pid)
                                )

                else:
                    idx = 0
                    fxn_string += "\n\trate = %s;" % rate_str
                    for sp in r._sources + r._dests:
                        s = sp()
                        # Get underlying rxd._IntracellularSpecies for the grid_id
                        if isinstance(s, species.Parameter) or isinstance(
                            s, species.ParameterOnRegion
                        ):
                            idx += 1
                            continue
                        if isinstance(s, species.Species):
                            s = s._intracellular_instances[reg]
                        elif isinstance(s, species.SpeciesOnRegion):
                            if s._region() in s._species()._intracellular_instances:
                                s = s._species()._intracellular_instances[s._region()]
                            else:
                                continue
                        if s._grid_id in ics_grid_ids:
                            operator = "+="
                        else:
                            operator = "="
                            ics_grid_ids.append(s._grid_id)
                        pid = [
                            pid
                            for pid, gid in enumerate(all_ics_gids)
                            if gid == s._grid_id
                        ][0]
                        fxn_string += "\n\trhs[%d] %s (%s)*rate;" % (
                            pid,
                            operator,
                            r._mult[idx],
                        )
                        idx += 1
            fxn_string += "\n}\n"
            for i, ele in enumerate(mults):
                if ele == []:
                    mults[i] = numpy.ones(len(reg._xs))
            mults = list(itertools.chain.from_iterable(mults))
            ics_register_reaction(
                0,
                len(all_ics_gids),
                len(ics_param_gids),
                _list_to_cint_array(all_ics_gids + ics_param_gids),
                numpy.asarray(mc3d_indices_start, dtype=numpy.uint64),
                mc3d_region_size,
                numpy.asarray(mults),
                _c_compile(fxn_string),
            )
    # Setup extracellular reactions
    if len(ecs_regions_inv) > 0:
        for reg in ecs_regions_inv:
            grid_ids = []
            all_gids = set()
            param_gids = set()
            fxn_string = _c_headers
            # TODO: find the nrn include path in python
            # It is necessary for a couple of function in python that are not in math.h
            fxn_string += (
                "void reaction(double* species_3d, double* params_3d, double* rhs)\n{"
            )
            # declare the "rate" variable if any reactions (non-rates)
            for rptr in [r for rlist in list(ecs_regions_inv.values()) for r in rlist]:
                if not isinstance(rptr(), rate.Rate):
                    fxn_string += "\n\tdouble rate;"
                    break
            # get a list of all grid_ids involved
            for s in ecs_species_by_region[reg]:
                if isinstance(s, species.Parameter) or isinstance(
                    s, species.ParameterOnExtracellular
                ):
                    sp = s[reg] if isinstance(s, species.Species) else s
                    param_gids.add(
                        sp._extracellular()._grid_id
                        if isinstance(sp, species.SpeciesOnExtracellular)
                        else sp._grid_id
                    )
                else:
                    sp = s[reg] if isinstance(s, species.Species) else s
                    all_gids.add(
                        sp._extracellular()._grid_id
                        if isinstance(sp, species.SpeciesOnExtracellular)
                        else sp._grid_id
                    )
            all_gids = list(all_gids)
            param_gids = list(param_gids)
            for rptr in ecs_regions_inv[reg]:
                r = rptr()
                rate_str = re.sub(
                    r"species_3d\[(\d+)\]",
                    lambda m: "species_3d[%i]"
                    % [
                        pid
                        for pid, gid in enumerate(all_gids)
                        if gid == int(m.groups()[0])
                    ][0],
                    r._rate_ecs[reg][-1],
                )
                rate_str = re.sub(
                    r"params_3d\[(\d+)\]",
                    lambda m: "params_3d[%i]"
                    % [
                        pid
                        for pid, gid in enumerate(param_gids)
                        if gid == int(m.groups()[0])
                    ][0],
                    rate_str,
                )
                if isinstance(r, rate.Rate):
                    s = r._species()
                    # Get underlying rxd._ExtracellularSpecies for the grid_id
                    if isinstance(s, species.Parameter) or isinstance(
                        s, species.ParameterOnExtracellular
                    ):
                        continue
                    elif isinstance(s, species.Species):
                        s = s[reg]._extracellular()
                    elif isinstance(s, species.SpeciesOnExtracellular):
                        s = s._extracellular()
                    if s._grid_id in grid_ids:
                        operator = "+="
                    else:
                        operator = "="
                        grid_ids.append(s._grid_id)
                    pid = [
                        pid for pid, gid in enumerate(all_gids) if gid == s._grid_id
                    ][0]
                    fxn_string += "\n\trhs[%d] %s %s;" % (pid, operator, rate_str)
                else:
                    idx = 0
                    fxn_string += "\n\trate = %s;" % rate_str
                    for sp in r._sources + r._dests:
                        s = sp()
                        # Get underlying rxd._ExtracellularSpecies for the grid_id
                        if isinstance(s, species.Parameter) or isinstance(
                            s, species.ParameterOnExtracellular
                        ):
                            idx += 1
                            continue
                        if isinstance(s, species.Species):
                            s = s[reg]._extracellular()
                        elif isinstance(s, species.SpeciesOnExtracellular):
                            s = s._extracellular()
                        if s._grid_id in grid_ids:
                            operator = "+="
                        else:
                            operator = "="
                            grid_ids.append(s._grid_id)
                        pid = [
                            pid for pid, gid in enumerate(all_gids) if gid == s._grid_id
                        ][0]
                        fxn_string += "\n\trhs[%d] %s (%s)*rate;" % (
                            pid,
                            operator,
                            r._mult[idx],
                        )
                        idx += 1
            fxn_string += "\n}\n"
            ecs_register_reaction(
                0,
                len(all_gids),
                len(param_gids),
                _list_to_cint_array(all_gids + param_gids),
                _c_compile(fxn_string),
            )


def _init():
    if len(species._all_species) == 0:
        return None
    initializer._do_init()
    # TODO: check about the 0<x<1 problem alluded to in the documentation
    h.define_shape()

    # if the shape has changed update the nodes
    _update_node_data()

    if species._has_1d:
        section1d._purge_cptrs()

    for sr in _species_get_all_species():
        s = sr()
        if s is not None:
            # TODO: are there issues with hybrid or 3D here? (I don't think so, but here's a bookmark just in case)
            s._register_cptrs()
            s._finitialize()
    _setup_matrices()
    # if species._has_1d and species._1d_submatrix_n():
    # volumes = node._get_data()[0]
    # zero_volume_indices = (numpy.where(volumes == 0)[0]).astype(numpy.int_)
    # setup_solver(_node_get_states(), len(_node_get_states()), zero_volume_indices, len(zero_volume_indices), h._ref_t, h._ref_dt)
    clear_rates()
    _setup_memb_currents()
    _compile_reactions()


def _include_flux(force=False):
    from .node import _node_fluxes
    from . import node

    if force or node._has_node_fluxes:
        index1D = []
        source1D = []
        scale1D = []
        grids = dict()
        for idx, t, src, sc, rptr in zip(
            _node_fluxes["index"],
            _node_fluxes["type"],
            _node_fluxes["source"],
            _node_fluxes["scale"],
            _node_fluxes["region"],
        ):
            if t == -1:
                index1D.append(idx)
                source1D.append(src)
                scale1D.append(sc * node._volumes[idx])
            else:
                gid = t
                if gid not in grids:
                    grids[gid] = {"index": [], "source": [], "scale": []}
                grids[gid]["index"].append(idx)
                grids[gid]["source"].append(src)
                grids[gid]["scale"].append(sc * rptr().volume(idx))
        counts3D = []
        grids3D = sorted(grids.keys())
        index3D = []
        source3D = []
        scale3D = []
        for gid in grids3D:
            counts3D.append(len(grids[gid]["index"]))
            index3D.extend(grids[gid]["index"])
            source3D.extend(grids[gid]["source"])
            scale3D.extend(grids[gid]["scale"])
        rxd_include_node_flux1D(
            len(index1D),
            _list_to_clong_array(index1D),
            _list_to_cdouble_array(scale1D),
            _list_to_pyobject_array(source1D),
        )

        rxd_include_node_flux3D(
            len(grids3D),
            _list_to_cint_array(counts3D),
            _list_to_cint_array(grids3D),
            _list_to_clong_array(index3D),
            _list_to_cdouble_array(scale3D),
            _list_to_pyobject_array(source3D),
        )
        node._has_node_fluxes = False


def _init_concentration():
    if len(species._all_species) == 0:
        return None
    for sr in _species_get_all_species():
        s = sr()
        if s is not None:
            # TODO: are there issues with hybrid or 3D here? (I don't think so, but here's a bookmark just in case)
            s._finitialize()


_has_nbs_registered = False
_nbs = None
do_setup_matrices_fptr = None
do_setup_units_fptr = None


def _do_nbs_register():
    global _has_nbs_registered, _nbs, _fih, _fih2, _fih3, do_setup_matrices_fptr, do_setup_units_fptr

    if not _has_nbs_registered:
        # from neuron import nonvint_block_supervisor as _nbs

        _has_nbs_registered = True
        # _nbs.register(_callbacks)  not used

        #
        # register the initialization handler and the ion register handler
        #
        _fih = h.FInitializeHandler(_init_concentration)
        _fih3 = h.FInitializeHandler(3, _init)

        set_setup_matrices = nrn_dll_sym("set_setup_matrices")
        set_setup_matrices.argtypes = [fptr_prototype]
        do_setup_matrices_fptr = fptr_prototype(_setup_matrices)
        set_setup_matrices(do_setup_matrices_fptr)

        set_setup_units = nrn_dll_sym("set_setup_units")
        set_setup_units.argtypes = [fptr_prototype]
        do_setup_units_fptr = fptr_prototype(_setup_units)
        set_setup_units(do_setup_units_fptr)

        _fih2 = h.FInitializeHandler(3, initializer._do_ion_register)

        #
        # register scatter/gather mechanisms
        #
        _cvode_object.extra_scatter_gather(0, _after_advance)

        # register save state mechanism
        import neuron
        from neuron import rxd

        neuron.register_savestate(
            "rxd-9e015bfa-93ba-485c-9dd2-09857ae58e4d",
            rxd.save_state,
            rxd.restore_state,
        )


# register the Python callbacks
do_setup_fptr = fptr_prototype(_setup)
do_initialize_fptr = fptr_prototype(_init)
set_setup(do_setup_fptr)
set_initialize(do_initialize_fptr)


def _windows_remove_dlls():
    global _windows_dll_files, _windows_dll
    if _windows_dll != []:
        if hasattr(ctypes, "WinDLL"):
            kernel32 = ctypes.WinDLL("kernel32")
            kernel32.FreeLibrary.argtypes = [ctypes.c_void_p]
        else:
            kernel32 = ctypes.windll.kernel32
    for (dll_ptr, filepath) in zip(_windows_dll, _windows_dll_files):
        dll = dll_ptr()
        if dll:
            handle = dll._handle
            del dll
            kernel32.FreeLibrary(handle)
        try:
            os.remove(filepath)
        except:
            pass
    _windows_dll_files = []
    _windows_dll = []


def nthread(n=None):
    if n:
        _set_num_threads(n)
    return _get_num_threads()



def test_node_selection(neuron_instance):
    """Test selection of 3D nodes"""

    h, rxd, data, save_path = neuron_instance

    rxd.set_solve_type(dimension=3)
    dend = h.Section(name="dend")
    dend.L = 10
    dend.diam = 2

    cyt = rxd.Region([dend])
    c = rxd.Species(cyt)

    nodes = c.nodes

    assert nodes((5.2, 0.1, 0.3))[0].x3d == 5.125
    assert nodes((5.2, 0.1, 0.3))[0].y3d == 0.125
    assert nodes((5.2, 0.1, 0.3))[0].z3d == 0.375
    assert len(nodes((5, 0, 0))) == 1



import neuron
from neuron import h, nrn, hoc, nrn_dll_sym
from . import region, constants
from . import rxdsection
import numpy
import weakref
from .rxdException import RxDException
import warnings
import ctypes

from collections.abc import Callable


# function to change extracellular diffusion
set_diffusion = nrn_dll_sym("set_diffusion")
set_diffusion.argtypes = [
    ctypes.c_int,
    ctypes.c_int,
    numpy.ctypeslib.ndpointer(ctypes.c_double),
    ctypes.c_int,
]
set_diffusion.restype = ctypes.c_int

# data storage
_volumes = numpy.array([])
_surface_area = numpy.array([])
_diffs = numpy.array([])
_states = numpy.array([])
_node_fluxes = {"index": [], "type": [], "source": [], "scale": [], "region": []}
_has_node_fluxes = False

_point_indices = {}

# node data types
_concentration_node = 0
_molecule_node = 1


def _get_data():
    return (_volumes, _surface_area, _diffs)


def _get_states():
    return _states


def _allocate(num):
    """allocate storage for num more nodes, return the starting index

    Note: no guarantee is made of preserving previous _ref
    """
    global _volumes, _surface_area, _diffs, _states
    start_index = len(_volumes)
    total = start_index + num
    _volumes.resize(total, refcheck=False)
    _surface_area.resize(total, refcheck=False)
    _diffs.resize(total, refcheck=False)
    _states.resize(total, refcheck=False)
    return start_index


def _remove(start, stop):
    """delete old volumes, surface areas and diff values in from global arrays"""
    global _volumes, _surface_area, _diffs, _states, _node_fluxes, _has_node_fluxes
    # Remove entries that have to be recalculated
    dels = list(range(start, stop))
    _volumes = numpy.delete(_volumes, dels)
    _surface_area = numpy.delete(_surface_area, dels)
    _diffs = numpy.delete(_diffs, dels)
    _states = numpy.delete(_states, dels)

    # remove _node_flux
    newflux = {"index": [], "type": [], "source": [], "scale": [], "region": []}
    for (i, idx) in enumerate(_node_fluxes["index"]):
        if idx not in dels:
            for key in _node_fluxes:
                newflux[key].append(_node_fluxes[key][i])
    _node_fluxes = newflux
    _has_node_fluxes = _node_fluxes["index"] != []

    # remove _node_flux
    for (i, idx) in enumerate(_node_fluxes["index"]):
        if idx in dels:
            for lst in _node_fluxes.values():
                del lst[i]


def _replace(old_offset, old_nseg, new_offset, new_nseg):
    """delete old volumes, surface areas and diff values in from global arrays
    move states so that the new segment value is equal to the old segment
    value that contains its centre"""
    global _volumes, _surface_area, _diffs, _states
    # remove entries that have to be recalculated
    start = old_offset
    stop = start + old_nseg + 1

    dels = list(range(start, stop))
    _volumes = numpy.delete(_volumes, dels)
    _surface_area = numpy.delete(_surface_area, dels)
    _diffs = numpy.delete(_diffs, dels)

    # replace states -- the new segment has the state from the old
    # segment which contains it's centre

    for j in range(new_nseg):
        i = int(((j + 0.5) / new_nseg) * old_nseg)
        _states[new_offset + j] = _states[old_offset + i]
    _states = numpy.delete(_states, list(range(start, stop)))

    # update _node_flux index
    for (i, idx) in enumerate(_node_fluxes["index"]):
        if idx in dels:
            j = int(((idx + 0.5) / new_nseg) * old_nseg)
            _node_fluxes["index"][i] = j


_numpy_element_ref = neuron.numpy_element_ref


class Node(object):
    def satisfies(self, condition):
        """Tests if a Node satisfies a given condition.

        If a nrn.Section object or RxDSection is provided, returns True if the Node lies in the section; else False.
        If a Region object is provided, returns True if the Node lies in the Region; else False.
        If a number between 0 and 1 is provided, returns True if the normalized position lies within the Node; else False.
        """
        if isinstance(condition, nrn.Section) or isinstance(
            condition, rxdsection.RxDSection
        ):
            return self._in_sec(condition)
        elif isinstance(condition, region.Region):
            return self.region == condition
        elif isinstance(condition, nrn.Segment):
            return self.segment == condition
        elif isinstance(condition, region.Extracellular):
            return self.region == condition
        raise RxDException("selector %r not supported for this node type" % condition)

    @property
    def _ref_concentration(self):
        """Returns a NEURON reference to the Node's concentration

        (The node must store concentration data. Use _ref_molecules for nodes
        storing molecule counts.)
        """
        # this points to rxd array only, will not change legacy concentration
        if self._data_type == _concentration_node:
            return self._ref_value
        else:
            raise RxDException(
                "_ref_concentration only available for concentration nodes"
            )

    @property
    def _ref_molecules(self):
        """Returns a NEURON reference to the Node's concentration

        (The node must store molecule counts. Use _ref_concentrations for nodes
        storing concentration.)
        """
        # this points to rxd array only, will not change legacy concentration
        if self._data_type == _molecule_node:
            return self._ref_value
        else:
            raise RxDException("_ref_molecules only available for molecule count nodes")

    @property
    def d(self):
        """Gets the diffusion rate within the compartment."""
        return _diffs[self._index]

    @d.setter
    def d(self, value):
        """Sets the diffusion rate within the compartment."""
        from . import rxd

        # TODO: make invalidation work so don't need to redo the setup each time
        # rxd._invalidate_matrices()
        _diffs[self._index] = value
        rxd._setup_matrices()

    @property
    def concentration(self):
        """Gets the concentration at the Node."""
        # TODO: don't use an if statement here... put the if statement at node
        #       construction and change the actual function that is pointed to
        if self._data_type == _concentration_node:
            return self.value
        else:
            # TODO: make this return a concentration instead of an error
            raise RxDException(
                "concentration property not yet supported for non-concentration nodes"
            )

    @concentration.setter
    def concentration(self, value):
        """Sets the concentration at the Node"""
        # TODO: don't use an if statement here... put the if statement at node
        #       construction and change the actual function that is pointed to
        if self._data_type == _concentration_node:
            self.value = value
        else:
            # TODO: make this set a concentration instead of raise an error
            raise RxDException(
                "concentration property not yet supported for non-concentration nodes"
            )

    @property
    def molecules(self):
        """Gets the molecule count at the Node."""
        # TODO: don't use an if statement here... put the if statement at node
        #       construction and change the actual function that is pointed to
        if self._data_type == _molecule_node:
            return self.value
        else:
            # TODO: make this return a molecule count instead of an error
            raise RxDException(
                "molecules property not yet supported for non-concentration nodes"
            )

    @molecules.setter
    def molecules(self, value):
        """Sets the molecule count at the Node"""
        # TODO: don't use an if statement here... put the if statement at node
        #       construction and change the actual function that is pointed to
        if self._data_type == _molecule_node:
            self.value = value
        else:
            # TODO: make this set a molecule count instead of raise an error
            raise RxDException(
                "molecules property not yet supported for non-concentration nodes"
            )

    @property
    def value(self):
        """Gets the value associated with this Node."""
        return _states[self._index]

    @value.setter
    def value(self, v):
        """Sets the value associated with this Node.

        For Species nodes belonging to a deterministic simulation, this is a concentration.
        For Species nodes belonging to a stochastic simulation, this is the molecule count.
        """
        _states[self._index] = v

    @property
    def _ref_value(self):
        """Returns a NEURON reference to the Node's value"""
        return _numpy_element_ref(_states, self._index)

    def include_flux(self, *args, **kwargs):
        """Include a flux contribution to a specific node.

        The flux can be described as a NEURON reference, a point process and a
        property, a Python function, or something that evaluates to a constant
        Python float.

        Supported units:
            molecule/ms
            mol/ms
            mmol/ms == millimol/ms == mol/s

        Examples:
            node.include_flux(mglur, 'ip3flux')           # default units: molecule/ms
            node.include_flux(mglur, 'ip3flux', units='mol/ms') # units: moles/ms
            node.include_flux(mglur._ref_ip3flux, units='molecule/ms')
            node.include_flux(lambda: mglur.ip3flux)
            node.include_flux(lambda: math.sin(h.t))
            node.include_flux(47)

        Warning:
            Flux denotes a change in *mass* not a change in concentration.
            For example, a metabotropic synapse produces a certain amount of
            substance when activated. The corresponding effect on the node's
            concentration depends on the volume of the node. (This scaling is
            handled automatically by NEURON's rxd module.)
        """
        global _has_node_fluxes, _node_fluxes
        if len(args) not in (1, 2):
            raise RxDException("node.include_flux takes only one or two arguments")
        if "units" in kwargs:
            units = kwargs.pop("units")
        else:
            units = "molecule/ms"
        if len(kwargs):
            raise RxDException("Unknown keyword arguments: %r" % list(kwargs.keys()))
        # take the value, divide by scale to get mM um^3
        # once this is done, we need to divide by volume to get mM
        # TODO: is division still slower than multiplication? Switch to mult.
        if units == "molecule/ms":
            scale = constants.molecules_per_mM_um3()
        elif units == "mol/ms":
            # You have: mol
            # You want: (millimol/L) * um^3
            #    * 1e+18
            #    / 1e-18
            scale = 1e-18
        elif units in ("mmol/ms", "millimol/ms", "mol/s"):
            # You have: millimol
            # You want: (millimol/L)*um^3
            #    * 1e+15
            #    / 1e-15
            scale = 1e-15
        else:
            raise RxDException("unknown unit: %r" % units)

        if len(args) == 1 and isinstance(args[0], hoc.HocObject):
            source = args[0]
            flux_type = 1
            try:
                # just a test access
                source[0]
            except:
                raise RxDException("HocObject must be a pointer")
        elif len(args) == 1 and isinstance(args[0], Callable):
            flux_type = 2
            source = args[0]
            warnings.warn(
                "Adding a python callback may slow down execution. Consider using a Rate and Parameter."
            )
        elif len(args) == 2:
            flux_type = 1
            try:
                source = getattr(args[0], "_ref_" + args[1])
            except:
                raise RxDException("Invalid two parameter form")

            # TODO: figure out a units checking solution that works
            # source_units = h.units(source)
            # if source_units and source_units != units:
            #    warnings.warn('Possible units conflict. NEURON says %r, but specified as %r.' % (source_units, units))
        else:
            success = False
            if len(args) == 1:
                try:
                    f = float(args[0])
                    source = f
                    flux_type = 3
                    success = True
                except:
                    pass
            if not success:
                raise RxDException("unsupported flux form")
        _node_fluxes["index"].append(self._index)
        if isinstance(self, Node1D):
            _node_fluxes["type"].append(-1)
        else:
            _node_fluxes["type"].append(self._grid_id)
        _node_fluxes["source"].append(source)
        _node_fluxes["scale"].append(scale)
        if isinstance(self, Node1D):
            _node_fluxes["region"].append(None)
        else:
            _node_fluxes["region"].append(weakref.ref(self._r))
        from .rxd import _structure_change_count

        _structure_change_count.value += 1
        _has_node_fluxes = True

    @value.getter
    def _state_index(self):
        return self._index


class Node1D(Node):
    def __init__(self, sec, i, location, data_type=_concentration_node):
        """n = Node1D(sec, i, location)
        Description:

        Constructs a Node object. These encapsulate multiple properties of a given
        reaction-diffusion compartment -- volume, surface area, concentration, etc... --
        but this association is provided only as a convenience to the user; all the data
        exists independently of the Node object.

        These objects are typically constructed by the reaction-diffusion code and
        provided to the user.

        Parameters:

        sec  -- the RxDSection containing the Node

        i -- the offset into the RxDSection's data

        location -- the location of the compartment within the section. For @aSection1D objects, this is the normalized position 0 <= x <= 1
        """
        self._sec = sec
        self._location = location
        self._index = i + sec._offset
        self._loc3d = None
        self._data_type = data_type

    def _update_loc3d(self):
        sec = self._sec
        length = sec.L
        normalized_arc3d = [sec._sec.arc3d(i) / length for i in range(sec._sec.n3d())]
        x3d = [sec._sec.x3d(i) for i in range(sec._sec.n3d())]
        y3d = [sec._sec.y3d(i) for i in range(sec._sec.n3d())]
        z3d = [sec._sec.z3d(i) for i in range(sec._sec.n3d())]
        loc1d = self._location
        self._loc3d = (
            numpy.interp(loc1d, normalized_arc3d, x3d),
            numpy.interp(loc1d, normalized_arc3d, y3d),
            numpy.interp(loc1d, normalized_arc3d, z3d),
        )

    def satisfies(self, condition):
        """Tests if a Node satisfies a given condition.

        If a nrn.Section object or RxDSection is provided, returns True if the Node lies in the section; else False.
        If a Region object is provided, returns True if the Node lies in the Region; else False.
        If a number between 0 and 1 is provided, returns True if the normalized position lies within the Node; else False.
        """
        try:
            # test Node conditions
            return super(Node1D, self).satisfies(condition)
        except:
            pass  # ignore the error and try Node1D specific conditions
        try:
            if 0 <= condition <= 1:
                dx = 1.0 / self._sec.nseg
                check_index = int(condition * self._sec.nseg)
                if check_index >= self._sec.nseg:
                    # if here, then at the 1 end
                    check_index = self._sec.nseg - 1
                # self_index should be unique (no repeats due to roundoff error) because the inside should always be 0.5 over an integer
                self_index = int(self._location * self._sec.nseg)
                return check_index == self_index

        except:
            pass
        raise RxDException("unrecognized node condition: %r" % condition)

    @property
    def x3d(self):
        """x coordinate"""
        if self._loc3d is None:
            self._update_loc3d()
        return self._loc3d[0]

    @property
    def y3d(self):
        """y coordinate"""
        if self._loc3d is None:
            self._update_loc3d()
        return self._loc3d[1]

    @property
    def z3d(self):
        """z coordinate"""
        if self._loc3d is None:
            self._update_loc3d()
        return self._loc3d[2]

    @property
    def volume(self):
        """The volume of the compartment in cubic microns.

        Read only."""
        from . import rxd

        rxd._update_node_data()
        return _volumes[self._index]

    @property
    def segment(self):
        return self._sec._sec(self.x)

    @property
    def surface_area(self):
        """The surface area of the compartment in square microns.

        This is the area (if any) of the compartment that lies on the plasma membrane
        and therefore is the area used to determine the contribution of currents (e.g. ina) from
        mod files or kschan to the compartment's concentration.

        Read only.
        """
        from . import rxd

        rxd._update_node_data()
        return _surface_area[self._index]

    @property
    def x(self):
        """The normalized position of the center of the compartment.

        Read only."""
        # TODO: will probably want to change this to be more generic for higher dimensions
        return self._location

    @property
    def region(self):
        """The region containing the compartment."""
        return self._sec._region

    @property
    def sec(self):
        """The Section containing the compartment."""
        return self._sec._sec

    def _in_sec(self, sec):
        return sec == self.sec or sec == self._sec

    @property
    def species(self):
        """The Species whose concentration is recorded at this Node."""
        return self._sec._species()


class Node3D(Node):
    def __init__(
        self, index, i, j, k, r, d, seg, speciesref, data_type=_concentration_node
    ):
        """
        Parameters
        ----------

        index : int
            the offset into the global rxd data
        i : int
            the x coordinate in the region's matrix
        j : int
            the y coordinate in the region's matrix
        k : int
            the z coordinate in the region's matrix
        r : rxd.Region
            the region that contains this node
        seg : nrn.Segment
            the segment containing this node
        """
        self._index = index
        self._i = i
        self._j = j
        self._k = k
        self._neighbors = None
        # TODO: store region as a weakref! (weakref.proxy?)
        self._r = r
        self._sec = h.SectionList([seg.sec])
        self._x = seg.x
        self._speciesref = speciesref
        self._data_type = data_type

        _point_indices.setdefault(self._r, {})
        _point_indices[self._r][(self._i, self._j, self._k)] = self._index

    @property
    def _seg(self):
        return list(self._sec)[0](self._x) if list(self._sec) else None

    def _find_neighbors(self):
        self._pos_x_neighbor = _point_indices[self._r].get(
            (self._i + 1, self._j, self._k)
        )
        self._neg_x_neighbor = _point_indices[self._r].get(
            (self._i - 1, self._j, self._k)
        )
        self._pos_y_neighbor = _point_indices[self._r].get(
            (self._i, self._j + 1, self._k)
        )
        self._neg_y_neighbor = _point_indices[self._r].get(
            (self._i, self._j - 1, self._k)
        )
        self._pos_z_neighbor = _point_indices[self._r].get(
            (self._i, self._j, self._k + 1)
        )
        self._neg_z_neighbor = _point_indices[self._r].get(
            (self._i, self._j, self._k - 1)
        )

        self._neighbors = (
            self._pos_x_neighbor,
            self._neg_x_neighbor,
            self._pos_y_neighbor,
            self._neg_y_neighbor,
            self._pos_z_neighbor,
            self._neg_z_neighbor,
        )

        return self._neighbors

    @property
    def neighbors(self):
        return (
            self._neighbors if self._neighbors is not None else self._find_neighbors()
        )

    @property
    def _grid_id(self):
        return self._speciesref()._intracellular_instances[self._r]._grid_id

    @property
    def surface_area(self):
        """The surface area of the compartment in square microns.

        This is the area (if any) of the compartment that lies on the plasma membrane
        and therefore is the area used to determine the contribution of currents (e.g. ina) from
        mod files or kschan to the compartment's concentration.

        Read only.
        """
        # TODO: should I have the commented out line?
        # rxd._update_node_data()
        return self._r._sa[self._index]

    def satisfies(self, condition):
        """Tests if a Node satisfies a given condition.

        If a nrn.Section object or RxDSection is provided, returns True if the Node lies in the section; else False.
        If a Region object is provided, returns True if the Node lies in the Region; else False.
        If a tuple is provided of length 3, return True if the Node contains the (x, y, z) point; else False.

        Does not currently support numbers between 0 and 1.
        """
        try:
            # test the Node conditions
            return super(Node3D, self).satisfies(condition)
        except:
            pass  # ignore the error and try Node3D specific conditions

        if isinstance(condition, tuple) and len(condition) == 3:
            x, y, z = condition
            mesh = self._r._mesh_grid
            return (
                int((x - mesh["xlo"]) / mesh["dx"]) == self._i
                and int((y - mesh["ylo"]) / mesh["dy"]) == self._j
                and int((z - mesh["zlo"]) / mesh["dz"]) == self._k
            )
        # check for a position condition so as to provide a more useful error
        try:
            if 0 <= condition <= 1:
                # TODO: the trouble here is that you can't do this super-directly based on x
                #       the way to do this is to find the minimum and maximum x values contained in the grid
                #       the extra difficulty with that is to handle boundary cases correctly
                #       (to test, consider a section 1 node wide by 15 discretized pieces long, access at 1./15, 2./15, etc...)
                raise RxDException(
                    "selecting nodes by normalized position not yet supported for 3D nodes; see comments in source about how to fix this"
                )
        except:
            pass
        raise RxDException("unrecognized node condition: %r" % condition)

    @property
    def x3d(self):
        # TODO: need to modify this to work with 1d
        mesh = self._r._mesh_grid
        return mesh["xlo"] + (self._i + 0.5) * mesh["dx"]

    @property
    def y3d(self):
        # TODO: need to modify this to work with 1d
        mesh = self._r._mesh_grid
        return mesh["ylo"] + (self._j + 0.5) * mesh["dy"]

    @property
    def z3d(self):
        # TODO: need to modify this to work with 1d
        mesh = self._r._mesh_grid
        return mesh["zlo"] + (self._k + 0.5) * mesh["dz"]

    @property
    def x(self):
        # TODO: can we make this more accurate?
        return self._x

    @property
    def segment(self):
        return self._seg

    def _in_sec(self, sec):
        return sec == self.sec

    @property
    def sec(self):
        return list(self._sec)[0] if any(self._sec) else None

    @property
    def volume(self):
        return self._r._vol[self._index]

    @property
    def region(self):
        """The region containing the compartment."""
        return self._r

    @property
    def species(self):
        """The Species whose concentration is recorded at this Node."""
        return self._speciesref()

    @property
    def d(self):
        """Gets the value associated with this Node."""
        sp = self._speciesref()._intracellular_instances[self._r]
        if hasattr(sp, "_dgrid"):
            return sp._dgrid[self._index]
        return sp._d

    @d.setter
    def d(self, v):
        sp = self._speciesref()._intracellular_instances[self._r]
        if not hasattr(sp, "_dgrid"):
            sp._dgrid = numpy.ndarray((3, sp._nodes_length), dtype=float, order="C")
            sp._dgrid[:] = sp._d.reshape(3, -1)
            set_diffusion(0, self._grid_id, sp._dgrid, sp._nodes_length)
        sp._dgrid[:, self._index] = v

    @property
    def value(self):
        """Gets the value associated with this Node."""
        return self._speciesref()._intracellular_instances[self._r].states[self._index]

    @value.setter
    def value(self, v):
        """Sets the value associated with this Node.

        For Species nodes belonging to a deterministic simulation, this is a concentration.
        For Species nodes belonging to a stochastic simulation, this is the molecule count.
        """
        self._speciesref()._intracellular_instances[self._r].states[self._index] = v

    @property
    def _ref_value(self):
        """Returns a HOC reference to the Node's value"""
        return (
            self._speciesref()
            ._intracellular_instances[self._r]
            ._states._ref_x[self._index]
        )


class NodeExtracellular(Node):
    def __init__(self, index, i, j, k, speciesref, regionref):
        """
        Parameters
        ----------

        index : int
            the offset into the global rxd data
        i : int
            the x coordinate in the region's matrix
        j : int
            the y coordinate in the region's matrix
        k : int
            the z coordinate in the region's matrix
        """
        self._index = index
        self._i = i
        self._j = j
        self._k = k
        # TODO: store region as a weakref! (weakref.proxy?)
        self._speciesref = speciesref
        self._regionref = regionref
        # TODO: support _molecule_node data type
        self._data_type = _concentration_node

    @property
    def x3d(self):
        return self._regionref()._xlo + (self._i + 0.5) * self._regionref()._dx[0]

    @property
    def y3d(self):
        return self._regionref()._ylo + (self._j + 0.5) * self._regionref()._dx[1]

    @property
    def z3d(self):
        return self._regionref()._zlo + (self._k + 0.5) * self._regionref()._dx[2]

    @property
    def region(self):
        """The extracellular space containing the node."""
        return self._regionref()

    @property
    def _r(self):
        """The extracellular space containing the node."""
        return self._regionref()

    @property
    def d(self):
        """Gets the diffusion rate within the compartment."""
        return self._speciesref()._d

    @d.setter
    def d(self, value):
        """Sets the diffusion rate within the compartment."""
        from . import rxd

        warnings.warn(
            "Changing the diffusion coefficient for an extracellular node changes diffusion coefficients for the whole extracellular grid."
        )
        # TODO: Replace zero with Parallel_grids id (here an in insert call)
        if hasattr(value, "__len__"):
            set_diffusion(0, self._grid_id, numpy.array(value, dtype=float), 1)
        else:
            set_diffusion(0, self._grid_id, numpy.repeat(float(value), 3), 1)

    @property
    def value(self):
        """Gets the value associated with this Node."""
        return (
            self._speciesref().__getitem__(self._r).states3d[self._i, self._j, self._k]
        )

    @value.setter
    def value(self, v):
        """Sets the value associated with this Node.

        For Species nodes belonging to a deterministic simulation, this is a concentration.
        For Species nodes belonging to a stochastic simulation, this is the molecule count.
        """
        self._speciesref().__getitem__(self._r).states3d[self._i, self._j, self._k] = v

    @property
    def _ref_value(self):
        """Returns a HOC reference to the Node's value"""
        return (
            self._speciesref()
            .__getitem__(self._regionref())
            ._extracellular()
            ._states._ref_x[self._index]
        )

    @value.getter
    def _state_index(self):
        return self._index

    @property
    def volume(self):
        ecs = self._regionref()
        if numpy.isscalar(ecs.alpha):
            return numpy.prod(ecs._dx) * ecs.alpha
        else:
            return numpy.prod(ecs._dx) * ecs.alpha[self._i, self._j, self._k]

    @property
    def _grid_id(self):
        return self._speciesref()._extracellular_instances[self._r]._grid_id

    def satisfies(self, condition):
        """Tests if a Node satisfies a given condition.

        If a nrn.Section object or RxDSection is provided, returns True if the Node lies in the section; else False.
        If a Region object is provided, returns True if the Node lies in the Region; else False.
        If a tuple is provided of length 3, return True if the Node contains the (x, y, z) point; else False.
        """
        try:
            # test the Node conditions
            return super(NodeExtracellular, self).satisfies(condition)
        except:
            pass  # ignore the error and try NodeExtracellular specific conditions

        if isinstance(condition, tuple) and len(condition) == 3:
            x, y, z = condition
            r = self._regionref()
            return (
                int((x - r._xlo) / r._dx[0]) == self._i
                and int((y - r._ylo) / r._dx[1]) == self._j
                and int((z - r._zlo) / r._dx[2]) == self._k
            )
        raise RxDException("unrecognized node condition: %r" % condition)



import distutils.util
import os
from neuron import coreneuron, h


def test_version_macros():
    """
    Use the special version_macros.mod mechanism to test the
    NRN_VERSION_X(maj, min, pat) preprocessor macros in VERBATIM blocks.
    """
    # Get the CMake PACKAGE_VERSION variable, which should always be in
    # major.minor.patch format. This should be baked into C++ as part of the main
    # NEURON build, not as part of nrnivmodl.
    ver = tuple(int(x) for x in h.nrnversion(0).split(".", 2))

    pc = h.ParallelContext()
    s = h.Section()
    s.insert("VersionMacros")
    coreneuron.enable = bool(
        distutils.util.strtobool(os.environ.get("NRN_CORENEURON_ENABLE", "false"))
    )
    coreneuron.verbose = True
    h.CVode().cache_efficient(True)
    h.finitialize()
    pc.set_maxstep(10)
    pc.psolve(0.1)

    def t(name):
        vals = [getattr(seg.VersionMacros, name + "_result") for seg in s]
        assert len(vals) == 1
        return bool(vals[0])

    assert t("eq8_2_0") == (ver == (8, 2, 0))
    assert t("ne9_0_1") == (ver != (9, 0, 1))
    assert t("gt9_0_0") == (ver > (9, 0, 0))
    assert t("lt42_1_2") == (ver < (42, 1, 2))
    assert t("gteq10_4_7") == (ver >= (10, 4, 7))
    assert t("lteq8_1_0") == (ver <= (8, 1, 0))
    assert t("explicit_gteq8_2_0") == (ver >= (8, 2, 0))


if __name__ == "__main__":
    test_version_macros()



from neuron import h

try:
    from neuron.rxd import species

    have_rxd = True
except:
    have_rxd = False


def get_mech_names():
    _mech_names = []
    mt = h.MechanismType(0)
    mname = h.ref("")
    for i in range(int(mt.count())):
        mt.select(i)
        mt.selected(mname)
        _mech_names.append(mname[0])
    return _mech_names


def get_rangevars(mech_name):
    ms = h.MechanismStandard(mech_name)
    suffix = "_" + mech_name
    lensuffix = len(suffix)
    mname = h.ref("")
    result = []
    for i in range(int(ms.count())):
        ms.name(mname, i)
        my_mname = mname[0]
        if my_mname.endswith(suffix):
            my_mname = my_mname[:-lensuffix]
        result.append(
            {
                "type": "NEURONMECH",
                "mech": mech_name,
                "var": my_mname,
                "name": mech_name + "." + my_mname,
            }
        )
    return result


def _get_ptrs(plot_what):
    ptrs = []
    if plot_what["type"] == "NEURON":
        plot_what_var = "_ref_" + plot_what["var"]
        for sec in h.allsec():
            try:
                getattr(sec(0.5), plot_what_var)
                bad_section = False
            except NameError:
                bad_section = True
            except AttributeError:
                bad_section = True
            if not bad_section:
                ptrs += [getattr(seg, plot_what_var) for seg in sec]
            else:
                ptrs += [None] * sec.nseg
    elif plot_what["type"] == "SECVAR":
        plot_what_var = "_ref_" + plot_what["var"]
        for sec in h.allsec():
            ptrs += [getattr(sec, plot_what_var)] * sec.nseg
    elif plot_what["type"] == "NEURONMECH":
        plot_what_mech = plot_what["mech"]
        plot_what_var = "_ref_" + plot_what["var"]
        for sec in h.allsec():
            for seg in sec:
                try:
                    mech = getattr(seg, plot_what_mech)
                except:
                    ptrs.append(None)
                    continue
                ptrs.append(getattr(mech, plot_what_var))
    elif plot_what["type"] == "NEURONRXD":
        # TODO: there has to be something more efficient than this
        # TODO: this probably won't work well with 3D nodes
        sp = plot_what["species"]
        region = plot_what["region"]
        for s in species._all_species:
            s = s()
            if s is None:
                continue
            if s.name == sp:
                rs = s._regions
                if hasattr(rs, "__len__"):
                    rs = list(rs)
                else:
                    rs = list([rs])
                for r in rs:
                    if r._name == region:
                        base_obj = s[r].nodes
                        for sec in h.allsec():
                            sec_obj = base_obj(sec)
                            if sec_obj:
                                for seg in sec:
                                    nodes = sec_obj(seg.x)
                                    if len(nodes) != 1:
                                        print("gui: node length mismatch; 3D?")
                                    if not nodes:
                                        ptrs.append(None)
                                    else:
                                        # TODO: change this to use molecules when appropriate (stochastic sims)
                                        ptrs.append(nodes[0]._ref_concentration)
                        break
                else:
                    continue
                break
        else:
            # no such species found
            for sec in h.allsec():
                ptrs += [None] * sec.nseg

    elif plot_what["type"] is None:
        # no such species found
        for sec in h.allsec():
            ptrs += [None] * sec.nseg

    else:
        print("gui: unknown plot_what type", plot_what)

    return ptrs


def _neuron_var(name):
    return {"type": "NEURON", "var": name, "name": name}


def _section_var(name):
    return {"type": "SECVAR", "var": name, "name": name}


def get_rxd_vars(secs):
    # TODO: filter by sections present
    result = []
    if have_rxd:
        for sp in species._all_species:
            sp = sp()
            if sp is not None and sp.name is not None:
                regions = sp._regions
                if hasattr(regions, "__len__"):
                    regions = list(regions)
                else:
                    regions = list([regions])
                result += [
                    {
                        "type": "NEURONRXD",
                        "name": "%s[%s]" % (sp.name, r._name),
                        "species": sp.name,
                        "region": r._name,
                    }
                    for r in regions
                    if r._name is not None
                ]
    return result


def rangevars_present(secs):
    result = []
    for name in get_mech_names():
        for sec in secs:
            try:
                if hasattr(sec(0.5), name):
                    if name[-4:] != "_ion":
                        result += get_rangevars(name)
                    else:
                        ionname = name[:-4]
                        result += [
                            _neuron_var(name)
                            for name in [
                                "i" + ionname,
                                ionname + "i",
                                ionname + "o",
                                "e" + ionname,
                            ]
                        ]
                    break
            except NameError:
                pass
    result += get_rxd_vars(secs)
    return [{"type": None, "name": "No coloration"}] + sorted(
        [_neuron_var("cm"), _neuron_var("diam"), _neuron_var("i_cap"), _neuron_var("v")]
        + result,
        key=lambda n: n["name"].lower(),
    )  # _section_var('Ra')



"""
Batch implementation suitable for parallel or serial
for stimulus current i iterating over a range of values
run a simulation and report spike frequency
save i & corresponding f to a file
optionally plot fi curve
"""

from mpi4py import MPI
from neuron import h
import time

# need the standard run system (didn't have to do this before, because this
# is loaded whenever we import the gui)
h.load_file("stdrun.hoc")

# create ParallelContext instance here so it exists whenever we need it
pc = h.ParallelContext()

# Simulation parameters

h.tstop = 500  # ms, more than long enough for 15 spikes at ISI = 25 ms
AMP0 = 0.1  # nA -- minimum stimulus
D_AMP = 0.02  # nA -- stim increment between runs
NRUNS = 30

# model specification
from cell import Cell

cell = Cell()

# instrumentation

# experimental manipulations
stim = h.IClamp(cell.soma(0.5))
stim.delay = 1  # ms
stim.dur = 1e9
stim.amp = 0.1  # nA


def fstimamp(run_id):
    """takes run i_d, returns corresponding stimulus amplitude"""
    return AMP0 + run_id * D_AMP


def set_params(run_id):
    """sets stimulus amplitude to appropriate value for this run"""
    stim.amp = fstimamp(run_id)


# data recording and analysis

# count only those spikes that get to distal end of dend
nc = h.NetCon(cell.dend(1)._ref_v, None, sec=cell.dend)
nc.threshold = -10  # mV
spvec = h.Vector()
nc.record(spvec)

NSETTLE = 5  # ignore the first NSETTLE ISI (allow freq to stablize)
NINVL = 10  # num ISI from which frequency will be calculated
NMIN = NSETTLE + NINVL  # ignore recordings with fewer than this num of ISIs


def get_frequency(spvec):
    nspikes = spvec.size()
    if nspikes > NMIN:
        t2 = spvec[-1]  # last spike
        t1 = spvec[-(1 + NINVL)]  # NINVL prior to last spike
        return NINVL * 1.0e3 / (t2 - t1)
    else:
        return 0


#
# Simulation control
#


def fi(run_id):
    """set params, execute a simulation, analyze and return results"""
    # NB: run_id is sent via NEURON, which turns it into a float
    set_params(run_id)
    h.run()
    return (int(run_id), stim.amp, get_frequency(spvec))


# batch control

t_start = time.time()  # keep track of execution time past this point

# start execute loop on the workers
pc.runworker()

# code beyond this point is executed only by the master
# the master must now post jobs to the bulletin board


def batchrun(n):
    # preallocate lists of length n for storing parameters and results
    stims = [None] * n
    freqs = [None] * n
    for run_id in range(n):
        pc.submit(fi, run_id)
    count = 0
    while pc.working():
        run_id, amp, freq = pc.pyret()
        stims[run_id] = amp
        freqs[run_id] = freq
        count += 1
        print("Finished %d of %d." % (count, n))
    return stims, freqs


stims, freqs = batchrun(NRUNS)

pc.done()

# all simulations have been completed
# and the workers have been released
# but the boss still has things to do

#
# Reporting of results
#


def save_results(filename, stims, freqs):
    with open(filename, "w") as f:
        f.write("label:%s\n%d\n" % ("f", len(freqs)))
        for stim, freq in zip(stims, freqs):
            f.write("%g\t%g\n" % (stim, freq))


save_results("fi.dat", stims, freqs)

print("%g seconds" % (time.time() - t_start))  # report run time



from .testutils import compare_data, tol


def test_pure_diffusion(neuron_instance):
    """Test 1D diffusion in a single section"""

    h, rxd, data, save_path = neuron_instance
    dend = h.Section()
    dend.diam = 2
    dend.nseg = 101
    dend.L = 100

    diff_constant = 1

    r = rxd.Region([dend])
    ca = rxd.Species(
        r, d=diff_constant, initial=lambda node: 1 if 0.4 < node.x < 0.6 else 0
    )

    h.finitialize(-65)

    for t in [25, 50, 75, 100, 125]:
        h.continuerun(t)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_pure_diffusion_cvode(neuron_instance):
    """Test 1D diffusion in a single section with the variable step method."""

    h, rxd, data, save_path = neuron_instance
    dend = h.Section()
    dend.diam = 2
    dend.nseg = 101
    dend.L = 100

    diff_constant = 1

    r = rxd.Region(h.allsec())
    ca = rxd.Species(
        r, d=diff_constant, initial=lambda node: 1 if 0.4 < node.x < 0.6 else 0
    )

    # enable CVode and set atol
    h.CVode().active(1)
    h.CVode().atol(1e-6)

    h.finitialize(-65)

    for t in [25, 50, 75, 100, 125]:
        h.continuerun(t)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol



"""
Import this module if you would like to use the NEURON GUI.

It loads nrngui.hoc, and starts a thread to periodically process
the NEURON GUI event loop.
"""


from neuron import h

import threading
import time

# recursive, especially in case stop/start pairs called from doNotify code.
_lock = threading.RLock()


def stop():
    _lock.acquire()


def start():
    _lock.release()


def process_events():
    _lock.acquire()
    try:
        h.doNotify()
    except:
        print("Exception in gui thread")
    _lock.release()


class Timer:
    """Python version of the NEURON Timer class"""

    def __init__(self, callback):
        self._callable = callback
        self._interval = 1
        self._thread = None

    def seconds(self, interval):
        last_interval = interval
        self._interval = interval
        return last_interval

    def _do_callback(self):
        if callable(self._callable):
            self._callable()
        else:
            h(self._callable)
        if self._thread is not None:
            self.start()

    def start(self):
        self._thread = threading.Timer(self._interval, self._do_callback)
        self._thread.start()

    def end(self):
        if self._thread is not None:
            self._thread.cancel()
            self._thread = None


class LoopTimer(threading.Thread):
    """
    a Timer that calls f every interval
    """

    def __init__(self, interval, fun):
        """
        @param interval: time in seconds between call to fun()
        @param fun: the function to call on timer update
        """
        self.started = False
        self.interval = interval
        self.fun = fun
        threading.Thread.__init__(self)
        self.setDaemon(True)

    def run(self):
        h.nrniv_bind_thread(threading.current_thread().ident)
        self.started = True
        while True:
            self.fun()
            time.sleep(self.interval)


if (
    h.nrnversion(9) == "2"
    or h.nrnversion(8).find("mingw") > 0
    or h.nrnversion(8).find("Windows")
):
    timer = LoopTimer(0.1, process_events)
    timer.start()
    while not timer.started:
        time.sleep(0.001)

h.load_file("nrngui.hoc")

h.movie_timer = Timer("moviestep()")



from neuron import h

if not "NRN_ENABLE_MUSIC=ON" in h.nrnversion(6):
    print("skip music_tests; NRN_ENABLE_MUSIC is not ON")
    quit()

import subprocess

# the test system copies the files in the source folder to
# .../build/test/nrnmusic/music_tests/test/music_tests
# but launches python in .../build/test/nrnmusic/musictests
# there must be a way to avoid that, but for now I'm just
# going to hit it with a sledghammer.

import os

try:
    a = os.getcwd() + "/test/music_tests"
    os.chdir(a)
except:
    pass

# Also try to find path to music binary
my_env = os.environ.copy()
try:
    my_env["PATH"] = os.getenv("MUSIC_LIBDIR") + "/../bin:" + my_env["PATH"]
except:
    pass


def run(cmd):
    result = subprocess.run(cmd, shell=True, env=my_env, capture_output=True, text=True)
    result.check_returncode()
    return result


out2 = "numprocs=1\nRank 0: Event (0, 0.001175) detected at 0\nRank 0: Event (0, 0.013825) detected at 0\n"

out3 = "numprocs=1\nnumprocs=1\ngroup= 1  rank= 0\n1 0  call stdinit\n1 0  call psolve\n1 0  done\ngroup= 2  rank= 0\n2 0  call stdinit\n2 0  call psolve\n2 0  done\n"


def chkresult(r, std):
    # lines can be out of order
    rlines = sorted(result.stdout.split("\n"))
    stdlines = sorted(std.split("\n"))
    if rlines != stdlines:
        print("standard", std, "result", r)
    assert rlines == stdlines


result = run("mpirun -np 2 music test2.music")
chkresult(result, out2)

result = run("mpirun -np 2 music test3.music")
chkresult(result, out3)



from neuron import h, rxd

h.load_file("stdrun.hoc")
h.CVode().active(True)

sec = h.Section(name="sec")
sec.L = 10
sec.nseg = 11
sec.diam = 5
rxd.set_solve_type(dimension=3)

cyt = rxd.Region(h.allsec(), name="cyt", nrn_region="i")
ip3 = rxd.Species(
    cyt, name="ip3", initial=lambda nd: 1000 if nd.segment == sec(0.3) else 0
)


def callbackfun():
    return 1000


for nd in ip3.nodes(sec(0.1)):
    nd.include_flux(1000)

for nd in ip3.nodes(sec(0.5)):
    nd.include_flux(callbackfun)

for nd in ip3.nodes(sec(0.9)):
    nd.include_flux(sec(0.3)._ref_ip3i)

h.finitialize(-70)
h.CVode().event(10)
h.continuerun(10)
import numpy

ip33d = ip3._intracellular_instances[cyt]
con = []
for seg in sec:
    con.append(
        numpy.mean([nd.concentration * nd.volume for nd in ip3.nodes(seg)])
        / seg.volume()
    )
print(con)

con1D = [
    0.0,
    0.0009302760807252256,
    0.0,
    1000.0,
    0.0,
    0.0009302760807252256,
    0.0,
    0.0,
    0.0,
    0.000930276080725227,
    0.0,
]

diff = numpy.array(con) - numpy.array(con1D)



from neuron import *

h(
    """
begintemplate A
public x, s, o, xa, oa, f, p
strdef s
objref o, oa[2]
double xa[3]
proc init() { \
  x = $1 \
}
func f() { return $1*xa[$2] }
proc p() { x += 1 }
endtemplate A
"""
)

_cls = hclass(h.A)


class A1(_cls):
    def __new__(cls, arg):
        return _cls.__new__(cls, arg)

    def __init__(self, arg):  # note, arg used by h.A
        # self.bp = hoc.HocObject.baseattr(self, 'p')
        self.bp = self.baseattr("p")

    def p(self):
        self.bp()
        return self.x



"""test of pure diffusion

Robert A McDougal
March 2013 - June 2014
"""

from neuron import h, crxd as rxd
import numpy
import __main__

name = __main__.__file__
if name[-3:] == ".py":
    name = name[:-3]

h.load_file("stdrun.hoc")

dend = h.Section()
dend.diam = 2
dend.nseg = 101
dend.L = 100

diff_constant = 1

r = rxd.Region(h.allsec())
ca = rxd.Species(
    r, d=diff_constant, initial=lambda node: 1 if 0.4 < node.x < 0.6 else 0
)

h.finitialize()

if __name__ == "__main__":
    from matplotlib import pyplot

    for t in [25, 50, 75, 100, 125]:
        h.continuerun(t)
        pyplot.plot([seg.x for seg in dend], ca.states)

    pyplot.tight_layout()
    pyplot.savefig("{0}.png".format(name))
else:
    for t in [25, 50, 75, 100, 125]:
        h.continuerun(t)



from .rxdException import RxDException


class NodeList(list):
    def __init__(self, items):
        """Constructs a NodeList from items, a python iterable containing Node objects."""
        list.__init__(self, items)

    def __call__(self, restriction):
        """returns a sub-NodeList consisting of nodes satisfying restriction"""
        return NodeList([i for i in self if i.satisfies(restriction)])

    def __getitem__(self, key):
        if isinstance(key, slice):
            return NodeList(list.__getitem__(self, key))
        else:
            return list.__getitem__(self, key)

    def __getslice__(self, i, j):
        # Python 2 support for simple slicing
        return NodeList(list.__getslice__(self, i, j))

    @property
    def value(self):
        # TODO: change this when not everything is a concentration
        return self.concentration

    @value.setter
    def value(self, v):
        # TODO: change this when not everything is a concentration
        self.concentration = v

    @property
    def segment(self):
        return [node.segment for node in self]

    @property
    def concentration(self):
        """Returns the concentration of the Node objects in the NodeList as an iterable."""
        return [node.concentration for node in self]

    @concentration.setter
    def concentration(self, value):
        """Sets the concentration of the Node objects to either a constant or values based on an iterable."""
        if hasattr(value, "__len__"):
            if len(value) == len(self):
                for node, val in zip(self, value):
                    node.concentration = val
                return
            else:
                raise RxDException(
                    "concentration must either be a scalar or an iterable of the same length as the NodeList"
                )
        for node in self:
            node.concentration = value

    @property
    def _ref_value(self):
        if not self:
            raise RxDException("no nodes")
        if len(self) != 1:
            raise RxDException("node not unique")
        return self[0]._ref_value

    @property
    def _ref_concentration(self):
        if not self:
            raise RxDException("no nodes")
        if len(self) != 1:
            raise RxDException("node not unique")
        return self[0]._ref_concentration

    @property
    def diff(self):
        """Returns the diffusion constant of the Node objects in the NodeList as an iterable."""
        return [node.diff for node in self]

    @diff.setter
    def diff(self, value):
        """Sets the diffusion constant of the Node objects to either a constant or values based on an iterable."""
        if hasattr(value, "__len__"):
            if len(value) == len(self):
                for node, val in zip(self, value):
                    node.diff = val
                return
            else:
                raise RxDException(
                    "diff must either be a scalar or an iterable of the same length as the NodeList"
                )
        for node in self:
            node.diff = value

    def include_flux(self, *args, **kwargs):
        for node in self:
            node.include_flux(args, kwargs)

    def value_to_grid(self):
        """Returns a regular grid with the values of the 3d nodes in the list.

        The grid is a copy only.

        Grid points not belonging to the object are assigned a value of NaN.

        Nodes that are not 3d will be ignored. If there are no 3d nodes, returns
        a 0x0x0 numpy array.

        Warning: Currently only supports nodelists over 1 region.
        """
        import numpy
        from .node import Node3D

        # identify regions involved
        regions = set()
        for node in self:
            if isinstance(node, Node3D):
                regions.add(node.region)

        # default result that falls through if no regions
        result = numpy.zeros((0, 0, 0))

        if len(regions) > 1:
            # TODO: the reason for this restriction is the need to make sure
            #       the mesh lines up
            raise RxDException("value_to_grid currently only supports 1 region")

        for r in regions:
            # TODO: if allowing multiple regions, need to change this
            result = numpy.empty((max(r._xs) + 5, max(r._ys) + 5, max(r._zs) + 5))
            result.fill(numpy.nan)

            for node in self:
                if isinstance(node, Node3D):
                    # TODO: if allowing multiple regions, adjust i, j, k as needed
                    result[node._i, node._j, node._k] = node.value

        return result

    @property
    def volume(self):
        """An iterable of the volumes of the Node objects in the NodeList.

        Read only."""
        return [node.volume for node in self]

    @property
    def surface_area(self):
        """An iterable of the surface areas of the Node objects in the NodeList.

        Read only."""
        return [node.surface_area for node in self]

    @property
    def region(self):
        """An iterable of the Region objects corresponding to the Node objects in the NodeList.

        Read only."""
        return [node.region for node in self]

    @property
    def species(self):
        """An iterable of the Species objects corresponding to the Node objects in the NodeList.

        Read only."""
        return [node.species for node in self]

    @property
    def x(self):
        """An iterable of the normalized positions of the Node objects in the NodeList.

        Read only."""
        return [node.x for node in self]



from neuron import h

pc = h.ParallelContext()

# each subworld has 3 ranks if nhost_world is multiple of subsize
# (if not, last will have nworld%subsize ranks)
subsize = 3
pc.subworlds(subsize)

ibbs = pc.id_world() // subsize
nbbs = pc.nhost_world() // subsize
x = 0 if pc.nhost_world() % subsize == 0 else 1

if pc.id_world() == 0:
    print("id_world nhost_world id_bbs nhost_bbs ibbs nbbs  id  nhost")
pc.barrier()
print(
    "%5d %9d %8d %8d %7d %3d %5d %4d"
    % (
        pc.id_world(),
        pc.nhost_world(),
        pc.id_bbs(),
        pc.nhost_bbs(),
        ibbs,
        nbbs + x,
        pc.id(),
        pc.nhost(),
    )
)

assert pc.nhost() == (subsize if ibbs < nbbs else (pc.nhost_world() - subsize * nbbs))
assert pc.id() == pc.id_world() % subsize
# id_bbs and nhost_bbs for non-zero id not -1.
assert pc.nhost_bbs() == ((nbbs + x) if pc.id() == 0 else -1)
assert pc.id_bbs() == (ibbs if pc.id() == 0 else -1)


def f(arg):
    print("f(%s) id_world=%d id=%d" % (str(arg), pc.id_world(), pc.id()))
    return (arg, pc.id_world(), pc.id())


x = 0


def gcntxt(arg):
    global x
    x = int(arg)
    # print("gcntxt(%s) id_world=%d" % (str(int(arg)), pc.id_world()))


def gtest(arg):
    if x != int(arg):
        h.execerror("gtest %d != %d" % (x, int(arg)))
        pass
    # print("gtest x=%d id_world=%d" % (x, pc.id_world()))


pc.runworker()

for i in range(5):
    pc.submit(f, "submit %d" % i)
hs = h.ref("")
while pc.working():
    pc.upkstr(hs)
    print("working returned %s (submit passed %s)" % (str(pc.pyret()), hs[0]))

pc.context(gcntxt, 6)

for i in range(3 * pc.nhost_world()):
    pc.submit(gtest, 6)


x = 6
while pc.working():
    pass

pc.done()

h.quit()



from . import graphicsPrimitives as graphics
from .. import options


def find_voxel(x, y, z, g):
    """returns (i,j,k) of voxel containing point x,y,z if the point is within
    the grid, otherwise return the corresponding grid boundary.
    """
    # g is grid boundaries
    i = max(0, int((x - g["xlo"]) // g["dx"]))
    j = max(0, int((y - g["ylo"]) // g["dy"]))
    k = max(0, int((z - g["zlo"]) // g["dz"]))
    return (i, j, k)


def get_verts(voxel, g):
    """return list (len=8) of point coordinates (x,y,z) that are vertices of the voxel (i,j,k)"""
    (i, j, k) = voxel
    dx, dy, dz = g["dx"], g["dy"], g["dz"]
    v1_0, v1_1, v1_2 = g["xlo"] + i * dx, g["ylo"] + j * dy, g["zlo"] + k * dz
    vertices = [
        (v1_0, v1_1, v1_2),
        (v1_0 + dx, v1_1, v1_2),
        (v1_0 + dx, v1_1 + dy, v1_2),
        (v1_0, v1_1 + dy, v1_2),
        (v1_0, v1_1, v1_2 + dz),
        (v1_0 + dx, v1_1, v1_2 + dz),
        (v1_0 + dx, v1_1 + dy, v1_2 + dz),
        (v1_0, v1_1 + dy, v1_2 + dz),
    ]
    return vertices


def get_surr_rows(row, endpoints, g):
    """return list (len=4) of the rows surrounding the current one on all sides
    IF the surrounding row is within the bounds of the grid."""
    (y, z) = row
    surr = []
    if y >= 1:
        surr.append(((y - 1, z), endpoints))
    if z >= 1:
        surr.append(((y, z - 1), endpoints))
    if (g["ylo"] + (y + 1) * g["dy"]) < g["yhi"]:
        surr.append(((y + 1, z), endpoints))
    if (g["zlo"] + (z + 1) * g["dz"]) < g["zhi"]:
        surr.append(((y, z + 1), endpoints))
    return surr


def verts_in(f, voxel, surf, g):
    """return the number of vertices of this voxel that are contained within the surface"""
    verts = get_verts(voxel, g)
    ins = 0
    distlist = []
    for (x, y, z) in verts:
        if (
            g["xlo"] <= x <= g["xhi"]
            and g["ylo"] <= y <= g["yhi"]
            and g["zlo"] <= z <= g["zhi"]
        ):
            dist = f.distance(x, y, z)
        else:
            dist = float("inf")
        distlist.append(dist)
        if dist <= options.ics_distance_threshold:
            ins += 1
    if 1 <= ins <= 7:
        surf[voxel] = distlist
    return ins


def find_endpoints(f, surf, include_ga, row, guesses, g):
    """return the endpoints (L,R) contained in the frustum f; if only one voxel both endpoints will be the same; if none both will be None
    f: frustum object
    surf: surface voxels
    row: current row
    guesses: estimates for endpoints
    g: grid boundaries"""
    # +x or right endpoint
    Rend, Lend = None, None
    check_surf_L, check_surf_R = (None, None), (None, None)
    stop = False
    Ri = guesses[1]
    ogrverts = verts_in(f, (Ri, row[0], row[1]), surf, g)
    if ogrverts == 0:
        going_in = True
    elif 1 <= ogrverts < 8:
        going_in = False
        check_surf_R = (True, Ri)
    else:
        going_in = False
    while (0 <= Ri and (g["xlo"] + (Ri) * g["dx"]) < g["xhi"]) and not stop:
        verts = verts_in(f, (Ri, row[0], row[1]), surf, g)
        if verts == 0:
            if not going_in:
                stop = True
                continue
            else:
                if Ri == guesses[0]:
                    # row is empty between guesses
                    return (None, None)
                Ri -= 1
                continue
        elif verts == 8:
            Rend = Ri
            Ri += 1
            continue
        else:
            Rend = Ri
            if going_in:
                check_surf_R = (True, Ri)
                break
            Ri += 1

    # the -x or left endpoint
    stop = False
    Li = guesses[0]
    oglverts = verts_in(f, (Li, row[0], row[1]), surf, g)
    if oglverts == 0:
        going_in = True
    elif 1 <= oglverts < 8:
        going_in = False
        check_surf_L = (True, Li)
    else:
        going_in = False
    while (0 <= Li and (g["xlo"] + (Li) * g["dx"]) < g["xhi"]) and not stop:
        verts = verts_in(f, (Li, row[0], row[1]), surf, g)
        if verts == 0:
            if not going_in:
                stop = True
                continue
            else:
                # it's not empty or would have already returned
                Li += 1
                continue
        elif verts == 8:
            Lend = Li
            Li -= 1
            continue
        else:
            Lend = Li
            if going_in:
                check_surf_L = (True, Li)
                break
            Li -= 1

    # check for extra surface voxels missed
    if check_surf_R[0] and Lend is not None:
        r = check_surf_R[1]
        while r > Lend:
            verts = verts_in(f, (r, row[0], row[1]), surf, g)
            if verts == 8:
                break
            else:
                r -= 1

    if check_surf_L[0] and Rend is not None:
        l = check_surf_L[1]
        while l < Rend:
            verts = verts_in(f, (l, row[0], row[1]), surf, g)
            if verts == 8:
                break
            else:
                l += 1

    # if keeping non-surface but grid-adjacent voxels:
    if include_ga:
        surf.add((Lend, row[0], row[1]))
        surf.add((Rend, row[0], row[1]))

    return (Lend, Rend)


def voxelize(grid, Object, corners=None, include_ga=False):
    """return a list of all voxels (i,j,k) that contain part of the object
    Other returned elements: set of surface voxels, possibly_missed for error handling"""
    # include_ga is whether to include grid-adjacent voxels in the surface, even if entirely within the surface

    yes_voxels = set()
    checked = set()
    surface = {}

    if corners is not None:
        for i in range(4):
            x0, y0, z0 = corners[i][0], corners[i][1], corners[i][2]
            (i0, j0, k0) = find_voxel(x0, y0, z0, grid)

            # find the contained endpoints and start the set with initial row and initial endpoints
            s = set()
            ends = find_endpoints(Object, surface, include_ga, (j0, k0), (i0, i0), grid)
            if ends[0]:
                break

    else:
        if isinstance(Object, graphics.Sphere):
            x0, y0, z0 = Object.x, Object.y, Object.z
        else:
            x0, y0, z0 = Object._x0, Object._y0, Object._z0

        # find the starting voxel
        (i0, j0, k0) = find_voxel(x0, y0, z0, grid)
        # find the contained endpoints and start the set with initial row and initial endpoints
        s = set()
        ends = find_endpoints(
            Object, surface, include_ga, (j0, k0), (i0 - 1, i0 + 1), grid
        )

    # the given starting voxel is not actually found
    possibly_missed = False
    if not ends[0]:
        possibly_missed = True
        ends = (i0, i0)

    # ------

    for i in range(ends[0], ends[1] + 1):
        yes_voxels.add((i, j0, k0))
    # add that initial row to checked and the set (otherwise inital voxel missed)
    checked.add((j0, k0))
    s.add(((j0, k0), ends))
    while s:
        startr = s.pop()
        newr = get_surr_rows(startr[0], startr[1], grid)
        for r in newr:
            (row, guesses) = r
            if row not in checked:
                (Lend, Rend) = find_endpoints(
                    Object, surface, include_ga, row, guesses, grid
                )
                if Lend is not None:
                    for i in range(Lend, Rend + 1):
                        yes_voxels.add((i, row[0], row[1]))
                    s.add((row, (Lend, Rend)))
                checked.add(row)

    missed = False
    if possibly_missed and (
        len(yes_voxels) == 1
    ):  # no voxels were found, return empty set
        missed = (i0, j0, k0)
        yes_voxels = set()
    return [yes_voxels, surface, missed]



import sphinx.writers.html
import sphinx.ext.mathjax
import re


class HTMLTranslator(sphinx.writers.html.HTMLTranslator):
    def visit_displaymath(self, node):
        sphinx.ext.mathjax.html_visit_displaymath(self, node)

    def depart_displaymath(self, node):
        return

    def visit_math(self, node):
        sphinx.ext.mathjax.html_visit_math(self, node)

    def depart_math(self, node):
        return

    def depart_document(self, node):
        methods_by_class = {}
        for i, tag in enumerate(self.body):
            if re.match(r'<dl class="((py|hoc) )?method">', tag):
                method_id = self.body[i + 1].split("id=")[1].split('"')[1].split(".")
                method_name = method_id[-1]
                method_class = ".".join(method_id[:-1])
                if method_class not in methods_by_class:
                    methods_by_class[method_class] = []
                methods_by_class[method_class].append(method_name)
        class_names = set(methods_by_class.keys())
        for i, tag in enumerate(self.body):
            if re.match(r'<dl class="((py|hoc) )?class">', tag):
                class_names.add(self.body[i + 1].split("id=")[1].split('"')[1])
        functions = []
        for i, tag in enumerate(self.body):
            if re.match(r'<dl class="((py|hoc) )?function">', tag):
                try:
                    functions.append(self.body[i + 1].split("id=")[1].split('"')[1])
                except:
                    functions.append(self.body[i + 3])
        for i, tag in enumerate(self.body):
            if re.match(r'<dl class="((py|hoc) )?data">', tag):
                data_id = self.body[i + 1].split("id=")[1].split('"')[1].split(".")
                data_name = data_id[-1]
                data_class = ".".join(data_id[:-1])
                if data_class:
                    if data_class not in methods_by_class:
                        methods_by_class[data_class] = []
                    methods_by_class[data_class].append(data_name)
                else:
                    functions.append(data_name)

        # sort everything
        sort_key = lambda s: s.lower()
        functions = sorted(functions, key=sort_key)
        class_names = sorted(list(class_names), key=sort_key)
        for key, values in zip(methods_by_class.keys(), methods_by_class.values()):
            methods_by_class[key] = sorted(values, key=sort_key)

        # add any missing class names to methods_by_class
        for name in class_names:
            if name not in methods_by_class:
                methods_by_class[name] = []

        jump_table = []
        if functions:
            fn_jmps = [
                '<a href="#%s" title="Link to this definition">%s</a>' % (name, name)
                for name in functions
            ]
            jump_table += ["<p>", " &middot; ".join(fn_jmps), "</p>"]

        for cl in class_names:
            method_jmps = [
                '<a href="#%s.%s" title="Link to this definition">%s</a>'
                % (cl, name, name)
                for name in methods_by_class[cl]
            ]
            jump_table += [
                "<p>",
                '<dl class="docutils"><dt><a href="#%s" title="Link to this definition">%s</a></dt><dd>%s</dd></dl>'
                % (cl, cl, " &middot; ".join(method_jmps)),
                "</p>",
            ]

        self.body = jump_table + self.body
        sphinx.writers.html.HTMLTranslator.depart_document(self, node)
